import numpy as np


class LinearOT:
    """Transformer class for linear OT problem."""

    def __init__(self, rho: float = 1.0, reg: float = 1e-6):
        self.rho = rho
        self.reg = reg

    def _fit(self, xs, xt, ws, wt):
        return barycenter(xs, xt, self.rho, ws, wt, self.reg)

    def fit(self, xs, xt, ws=None, wt=None):
        """Fit linear OT problem."""
        self.As, self.bs, self.At, self.bt = self._fit(xs, xt, ws, wt)

        return self

    def fit_transform(self, xs, xt, ws=None, wt=None):
        """Fit and transform linear OT problem."""
        self.As, self.bs, self.At, self.bt = self._fit(xs, xt, ws, wt)

        return xs @ self.As + self.bs, xt @ self.At + self.bt


def psd_sqrt(A):
    """Square root of a positive semidefinite matrix A."""
    u, s, v = np.linalg.svd(A)
    return u @ np.diag(s**0.5) @ v


def compute_stats(x, w=None, reg=1e-12):
    """Compute mean, covariance, correlation, and variances from set of data points."""
    d = x.shape[1]
    n = x.shape[0]

    # if weights not given, make them all 1
    if w is None:
        w = np.ones(n)

    # reshape weights to make them easier to multiply
    w_reshaped = w.reshape(-1, 1)

    # mean of the distribution
    m = (x.T @ w_reshaped / np.sum(w_reshaped)).T

    # center points
    x_centered = x - m

    # covariance of each distribution
    cs = (x_centered * w_reshaped).T @ x_centered / np.sum(w)
    cs += np.eye(d) * reg

    # convert covariance to correlation
    variances = np.diag(cs)

    # standarddevs
    stdevs_inverse = (np.sqrt(variances).reshape(-1, 1)) ** -1

    # compute correlation
    corr = cs * (stdevs_inverse @ stdevs_inverse.T)

    res = {"mean": m, "covariance": cs, "correlation": corr, "variance": variances}

    return res


def compute_transformation(cs, ct, ms, mt, rho=1.0, reg=1e-6):
    """Given mean and covariance, compute transport plan."""
    n, ix = ms.shape
    d = cs.shape[0]

    # add regularization to make sure invertible
    cs += np.eye(d) * reg
    ct += np.eye(d) * reg

    # square roots
    cs_sqrt = psd_sqrt(cs)
    ct_sqrt = psd_sqrt(ct)

    cst = psd_sqrt(ct_sqrt @ cs @ ct_sqrt)

    # compute mean and cov of barycenter
    mb = (1 - rho) * ms + rho * mt

    term1 = (1 - rho) * np.eye(d) + rho * ct_sqrt @ np.linalg.pinv(cst) @ ct_sqrt
    term2 = cs
    cb = term1 @ term2 @ term1

    # transport plan for source
    cs_sqrt_pinv = np.linalg.pinv(cs_sqrt)
    As = cs_sqrt_pinv @ psd_sqrt(cs_sqrt @ cb @ cs_sqrt) @ cs_sqrt_pinv
    bs = mb - (As @ ms.T).T

    # transport plan for target
    ct_sqrt_pinv = np.linalg.pinv(ct_sqrt)
    At = ct_sqrt_pinv @ psd_sqrt(ct_sqrt @ cb @ ct_sqrt) @ ct_sqrt_pinv
    bt = mb - (At @ mt.T).T

    return As.T, bs, At.T, bt


def barycenter(
    xs,
    xt,
    rho=1.0,
    ws=None,
    wt=None,
    reg: float = 1e-6,
):
    """Wasserstein barycenter between two Gaussians."""
    # number of points
    d = xs.shape[1]
    ns = xs.shape[0]
    nt = xt.shape[0]

    # if weights not given, make them all 1
    if ws is None:
        ws = np.ones(ns)
    if wt is None:
        wt = np.ones(nt)

    # reshape weights to make them easier to multiply
    ws_reshaped = ws.reshape(-1, 1)
    wt_reshaped = wt.reshape(-1, 1)

    # mean of each distribution
    ms = (xs.T @ ws_reshaped / np.sum(ws_reshaped)).T
    mt = (xt.T @ wt_reshaped / np.sum(wt_reshaped)).T
    # ms = np.mean(xs, axis=0)
    # mt = np.mean(xt, axis=0)

    # center each set of points
    xs_centered = xs - ms
    xt_centered = xt - mt

    # covariance of each distribution
    cs = (xs_centered * ws_reshaped).T @ xs_centered / np.sum(ws)
    ct = (xt_centered * wt_reshaped).T @ xt_centered / np.sum(wt)

    cs += np.eye(d) * reg
    ct += np.eye(d) * reg
    # cs = np.cov(xs.T) + np.eye(d) * reg
    # ct = np.cov(xt.T) + np.eye(d) * reg

    # square roots
    cs_sqrt = psd_sqrt(cs)
    ct_sqrt = psd_sqrt(ct)

    cst = psd_sqrt(ct_sqrt @ cs @ ct_sqrt)

    # compute mean and cov of b,,arycenter
    mb = (1 - rho) * ms + rho * mt

    term1 = (1 - rho) * np.eye(d) + rho * ct_sqrt @ np.linalg.pinv(cst) @ ct_sqrt
    term2 = cs
    cb = term1 @ term2 @ term1

    # transport plan for source
    cs_sqrt_pinv = np.linalg.pinv(cs_sqrt)
    As = cs_sqrt_pinv @ psd_sqrt(cs_sqrt @ cb @ cs_sqrt) @ cs_sqrt_pinv
    bs = mb - (As @ ms.T).T

    # transport plan for target
    ct_sqrt_pinv = np.linalg.pinv(ct_sqrt)
    At = ct_sqrt_pinv @ psd_sqrt(ct_sqrt @ cb @ ct_sqrt) @ ct_sqrt_pinv
    bt = mb - (At @ mt.T).T

    return As.T, bs, At.T, bt



import copy

import numpy as np
import pandas as pd
from tqdm import tqdm


from .evaluate import compute_celltype_purity


def SEACells(
    ad,
    build_kernel_on: str,
    n_SEACells: int,
    use_gpu: bool = False,
    verbose: bool = True,
    n_waypoint_eigs: int = 10,
    n_neighbors: int = 15,
    convergence_epsilon: float = 1e-3,
    l2_penalty: float = 0,
    max_franke_wolfe_iters: int = 50,
    use_sparse: bool = False,
):
    """Core SEACells class.

    :param ad: (AnnData) annotated data matrix
    :param build_kernel_on: (str) key corresponding to matrix in ad.obsm which is used to compute kernel for metacells
                            Typically 'X_pca' for scRNA or 'X_svd' for scATAC
    :param n_SEACells: (int) number of SEACells to compute
    :param use_gpu: (bool) whether to use GPU for computation
    :param verbose: (bool) whether to suppress verbose program logging
    :param n_waypoint_eigs: (int) number of eigenvectors to use for waypoint initialization
    :param n_neighbors: (int) number of nearest neighbors to use for graph construction
    :param convergence_epsilon: (float) convergence threshold for Franke-Wolfe algorithm
    :param l2_penalty: (float) L2 penalty for Franke-Wolfe algorithm
    :param max_franke_wolfe_iters: (int) maximum number of iterations for Franke-Wolfe algorithm
    :param use_sparse: (bool) whether to use sparse matrix operations. Currently only supported for CPU implementation.

    See cpu.py or gpu.py for descriptions of model attributes and methods.
    """
    if use_sparse:
        assert (
            not use_gpu
        ), "Sparse matrix operations are only supported for CPU implementation."
        try:
            from . import cpu
        except ImportError:
            import cpu
        model = cpu.SEACellsCPU(
            ad,
            build_kernel_on,
            n_SEACells,
            verbose,
            n_waypoint_eigs,
            n_neighbors,
            convergence_epsilon,
            l2_penalty,
            max_franke_wolfe_iters,
        )

        return model

    if use_gpu:
        try:
            from . import gpu
        except ImportError:
            import gpu

        model = gpu.SEACellsGPU(
            ad,
            build_kernel_on,
            n_SEACells,
            verbose,
            n_waypoint_eigs,
            n_neighbors,
            convergence_epsilon,
            l2_penalty,
            max_franke_wolfe_iters,
        )

    else:
        try:
            from . import cpu_dense
        except ImportError:
            import cpu_dense
        model = cpu_dense.SEACellsCPUDense(
            ad,
            build_kernel_on,
            n_SEACells,
            verbose,
            n_waypoint_eigs,
            n_neighbors,
            convergence_epsilon,
            l2_penalty,
            max_franke_wolfe_iters,
        )

    return model


def sparsify_assignments(A, thresh, keep_above_percentile=95):
    """
    :param: A is a cell x SEACells assignment matrix
    """
    A = copy.deepcopy(A)
    A[A == 0] = np.nan
    mins = np.nanpercentile(A, keep_above_percentile, axis=1).reshape(-1, 1)
    A = np.nan_to_num(A)
    mins[mins>thresh] = thresh
    A[A < mins] = 0

    # Renormalize
    A = A / A.sum(1, keepdims=True)

    return A


def summarize_by_soft_SEACell(
    ad, A, celltype_label=None, summarize_layer="raw", minimum_weight: float = 0.05
):
    """Summary of soft SEACell assignment.

    Aggregates cells within each SEACell, summing over all raw data x assignment weight for all cells belonging to a
    SEACell. Data is un-normalized and pseudo-raw aggregated counts are stored in .layers['raw'].
    Attributes associated with variables (.var) are copied over, but relevant per SEACell attributes must be
    manually copied, since certain attributes may need to be summed, or averaged etc, depending on the attribute.
    The output of this function is an anndata object of shape n_metacells x original_data_dimension.

    @param ad: (sc.AnnData) containing raw counts for single-cell data
    @param A: (np.array) of shape n_SEACells x n_cells containing assignment weights of cells to SEACells
    @param celltype_label: (str) optionally provide the celltype label to compute modal celltype per SEACell
    @param summarize_layer: (str) key for ad.layers to find raw data. Use 'raw' to search for ad.raw.X
    @param minimum_weight: (float) minimum value below which assignment weights are zero-ed out. If all cell assignment
                            weights are smaller than minimum_weight, the 95th percentile weight is used.
    @return: aggregated anndata containing weighted expression for aggregated SEACells
    """
    import scanpy as sc
    from scipy.sparse import csr_matrix

    compute_seacell_celltypes = False
    if celltype_label is not None:
        if celltype_label not in ad.obs.columns:
            raise ValueError(f"Celltype label {celltype_label} not present in ad.obs")
        compute_seacell_celltypes = True

    if summarize_layer == "raw" and ad.raw is not None:
        data = ad.raw.X
    else:
        data = ad.layers[summarize_layer]

    A = sparsify_assignments(A.T, thresh=minimum_weight)

    seacell_expressions = []
    seacell_celltypes = []
    seacell_purities = []
    for ix in tqdm(range(A.shape[1])):
        cell_weights = A[:, ix]
        # Construct the SEACell expression using the
        seacell_exp = (
            data.multiply(cell_weights[:, np.newaxis]).toarray().sum(0)
            / cell_weights.sum()
        )
        seacell_expressions.append(seacell_exp)

        if compute_seacell_celltypes:
            # Compute the consensus celltype and the celltype purity
            cell_weights = pd.DataFrame(cell_weights)
            cell_weights.index = ad.obs_names
            purity = (
                cell_weights.join(ad.obs[celltype_label])
                .groupby(celltype_label)
                .sum()
                .sort_values(by=0, ascending=False)
            )
            purity = purity / purity.sum()
            celltype = purity.iloc[0]
            seacell_celltypes.append(celltype.name)
            seacell_purities.append(celltype.values[0])

    seacell_expressions = csr_matrix(np.array(seacell_expressions))
    seacell_ad = sc.AnnData(seacell_expressions, dtype=seacell_expressions.dtype)
    seacell_ad.var_names = ad.var_names
    seacell_ad.obs["Pseudo-sizes"] = A.sum(0)
    if compute_seacell_celltypes:
        seacell_ad.obs["celltype"] = seacell_celltypes
        seacell_ad.obs["celltype_purity"] = seacell_purities
    seacell_ad.var_names = ad.var_names
    return seacell_ad


def summarize_by_SEACell(
    ad, SEACells_label="SEACell", celltype_label=None, summarize_layer="raw"
):
    """Summary of SEACell assignment.

    Aggregates cells within each SEACell, summing over all raw data for all cells belonging to a SEACell.
    Data is unnormalized and raw aggregated counts are stored .layers['raw'].
    Attributes associated with variables (.var) are copied over, but relevant per SEACell attributes must be
    manually copied, since certain attributes may need to be summed, or averaged etc, depending on the attribute.
    The output of this function is an anndata object of shape n_metacells x original_data_dimension.
    :return: anndata.AnnData containing aggregated counts.

    """
    import scanpy as sc
    from scipy.sparse import csr_matrix

    # Set of metacells
    metacells = ad.obs[SEACells_label].unique()

    # Summary matrix
    summ_matrix = pd.DataFrame(0.0, index=metacells, columns=ad.var_names)

    for m in tqdm(summ_matrix.index):
        cells = ad.obs_names[ad.obs[SEACells_label] == m]
        if summarize_layer == "X":
            summ_matrix.loc[m, :] = np.ravel(ad[cells, :].X.sum(axis=0))
        elif summarize_layer == "raw" and ad.raw is not None:
            summ_matrix.loc[m, :] = np.ravel(ad[cells, :].raw.X.sum(axis=0))
        else:
            summ_matrix.loc[m, :] = np.ravel(
                ad[cells, :].layers[summarize_layer].sum(axis=0)
            )

    # Ann data

    # Counts
    meta_ad = sc.AnnData(csr_matrix(summ_matrix), dtype=csr_matrix(summ_matrix).dtype)
    meta_ad.obs_names, meta_ad.var_names = summ_matrix.index.astype(str), ad.var_names
    meta_ad.layers["raw"] = csr_matrix(summ_matrix)

    # Also compute cell type purity
    if celltype_label is not None:
        # TODO: Catch specific exception
        try:
            purity_df = compute_celltype_purity(ad, celltype_label)
            meta_ad.obs = meta_ad.obs.join(purity_df)
        except Exception as e:  # noqa: BLE001
            print(f"Cell type purity failed with Exception {e}")

    return meta_ad



from ..pp import *
import scanpy as sc
import numpy as np

from .._settings import add_reference,settings

def lazy(adata,
         species='human',
         reforce_steps=[],
         sample_key=None,
         qc_kwargs=None,
         preprocess_kwargs=None,
         pca_kwargs=None,
         harmony_kwargs=None,
         scvi_kwargs=None,
         ):
    """
    This is a very interesting function. We can use this function to avoid many unnecessary steps.

    arguments:
        adata: the data to analysis
        reforce_steps: we can reforce run lazy step, because some step have been run and will be skipped. 
                        ['qc','pca','preprocess','scaled','Harmony','scVI','eval_bench','eval_clusters']
        sample_key: the key store in `adata.obs` to batch correction.
    
    """
    mode=settings.mode
    print(f'🔧 The mode of lazy is {mode}')
    if mode == 'cpu-gpu-mixed':
        try:
            import pymde
        except:
            print('❌ pymde package not found, we will install it now')
            import pip
            pip.main(['install','pymde'])
            import pymde
    else:
        pass

    #step 0: check packages:
    try:
        import louvain
    except:
        print('❌ Louvain package not found, we will install it now')
        import pip
        pip.main(['install','louvain'])
        import louvain
    

    try:
        import louvain
    except:
        print('❌ louvain package not found, we will install it now')
        import pip
        pip.main(['install','louvain'])
        import louvain
        
    print('✅ All packages used in lazy are installed')
    
    #step 1: qc:
    adata.var_names_make_unique()
    adata.obs_names_make_unique()
    adata.copy()
    #check adata.uns['status']
    if 'status' not in adata.uns.keys():
        adata.uns['status'] = {}

    if  ('qc' not in adata.uns['status'].keys()) or ('qc' in adata.uns['status'].keys() and adata.uns['status']['qc'] == False) or ('qc' in reforce_steps):
        print('❌ QC step didn\'t start, we will start it now')
        if qc_kwargs is None:
            qc_kwargs = {
                'tresh': {'mito_perc': 0.2, 'nUMIs': 500, 'detected_genes': 250},
                'doublets_method': 'scrublet',
                'batch_key': sample_key
            }
        print(f'🔧 The argument of qc we set\n'
              f'   mito_perc: {qc_kwargs["tresh"]["mito_perc"]}\n'
              f'   nUMIs: {qc_kwargs["tresh"]["nUMIs"]}\n'
              f'   detected_genes: {qc_kwargs["tresh"]["detected_genes"]}\n'
              f'   doublets_method: {qc_kwargs["doublets_method"]}\n'
              f'   batch_key: {qc_kwargs["batch_key"]}\n'
              )
        adata = qc(adata,
                **qc_kwargs)
    else:
        print('✅ QC step already finished, skipping it')
        print(f'🔧 The argument of qc we set\n'
              f'   mito_perc: {adata.uns["status_args"]["qc"]["mito_perc"]}\n'
              f'   nUMIs: {adata.uns["status_args"]["qc"]["nUMIs"]}\n'
              f'   detected_genes: {adata.uns["status_args"]["qc"]["detected_genes"]}\n'
              f'   doublets_method: {adata.uns["status_args"]["qc"]["doublets_method"]}\n'
              f'   batch_key: {adata.uns["status_args"]["qc"]["batch_key"]}\n'
              )

    #step 2: normalization and highly variable genes:
    if  ('preprocess' not in adata.uns['status'].keys()) or ('preprocess' in adata.uns['status'].keys() and adata.uns['status']['preprocess'] == False)  or ('preprocess' in reforce_steps):
        print('❌ Preprocess step didn\'t start, we will start it now')
        if preprocess_kwargs is None:
            preprocess_kwargs = {
                'mode': 'shiftlog|pearson',
                'n_HVGs': 2000,
                'target_sum': 50*1e4
            }
        print(f'🔧 The argument of preprocess we set\n'
              f'   mode: {preprocess_kwargs["mode"]}\n'
              f'   n_HVGs: {preprocess_kwargs["n_HVGs"]}\n'
              f'   target_sum: {preprocess_kwargs["target_sum"]}\n'
              )
        adata = preprocess(adata,**preprocess_kwargs)
    else:
        print('✅ Preprocess step already finished, skipping it')
        print(f'🔧 The argument of preprocess in data\n'
              f'   mode: {adata.uns["status_args"]["preprocess"]["mode"]}\n'
              f'   n_HVGs: {adata.uns["status_args"]["preprocess"]["n_HVGs"]}\n'
              f'   target_sum: {adata.uns["status_args"]["preprocess"]["target_sum"]}\n'
              )
        
    if  ('scaled' not in adata.uns['status'].keys()) or ('scaled' in adata.uns['status'].keys() and adata.uns['status']['scaled'] == False)  or ('scaled' in reforce_steps):
        print('❌ Scaled step didn\'t start, we will start it now')
        scale(adata)
    else:
        print('✅ Scaled step already finished, skipping it')
    
    #step 3: PCA:
    if  ('pca' not in adata.uns['status'].keys()) or ('pca' in adata.uns['status'].keys() and adata.uns['status']['pca'] == False)  or ('pca' in reforce_steps):
        print('❌ PCA step didn\'t start, we will start it now')
        if sc.__version__>='1.11.0':
            if pca_kwargs is None:
                pca_kwargs = {
                    'layer':'scaled',
                    'n_pcs':50,
                    'use_highly_variable': True,
                }
            if ('highly_variable' not in adata.var.columns) and ('highly_variable_features' in adata.var.columns):
                adata.var['highly_variable'] = adata.var['highly_variable_features'].tolist()
            print(f'🔧 The argument of PCA we set\n'
                    f'   layer: {pca_kwargs["layer"]}\n'
                    f'   n_pcs: {pca_kwargs["n_pcs"]}\n'
                    f'   use_highly_variable: {pca_kwargs["use_highly_variable"]}\n'
                    )
            pca(adata,**pca_kwargs)
            adata.obsm['X_pca']=adata.obsm["scaled|original|X_pca"]
        else:
            print('❌ The version of scanpy is lower than 1.11.0, we will use the old version of PCA function (sc.pp.pca)')
            if pca_kwargs is None:
                pca_kwargs = {
                    'layer':'scaled',
                    'n_comps':50,
                    'use_highly_variable': True,
                }
            if ('highly_variable' not in adata.var.columns) and ('highly_variable_features' in adata.var.columns):
                adata.var['highly_variable'] = adata.var['highly_variable_features'].tolist()
            print(f'🔧 The argument of PCA we set\n'
                    f'   layer: {pca_kwargs["layer"]}\n'
                    f'   n_comps: {pca_kwargs["n_comps"]}\n'
                    f'   use_highly_variable: {pca_kwargs["use_highly_variable"]}\n'
                    )
            sc.pp.pca(adata,**pca_kwargs)
        print('❌ The version of scanpy is lower than 1.11.0, GPU mode will not work, we will use CPU mode')
        print('    If you want to use GPU mode, please update scanpy to 1.11.0 or higher')
    else:
        print('✅ PCA step already finished, skipping it')

    #step 4 Score cell cycle:
    if  ('cell_cycle' not in adata.uns['status'].keys()) or ('cell_cycle' in adata.uns['status'].keys() and adata.uns['status']['cell_cycle'] == False)  or ('cell_cycle' in reforce_steps):
        print('❌ Cell cycle scoring step didn\'t start, we will start it now')
        score_genes_cell_cycle(adata,species=species)
    else:
        print('✅ Cell cycle scoring step already finished, skipping it')

    #step 5 batch remove:
    adata_hvg=adata.copy()
    if 'highly_variable_features' in adata_hvg.var.columns:
        adata_hvg = adata_hvg[:, adata_hvg.var.highly_variable_features]
    elif 'highly_variable' in adata_hvg.var.columns:
        adata_hvg = adata_hvg[:, adata_hvg.var.highly_variable]
    from . import batch_correction
    if ('X_harmony' not in adata.obsm.keys()) or ('Harmony' in reforce_steps):
        print('❌ Batch Correction: `Harmony` step didn\'t start, we will start it now')
        if harmony_kwargs is None:
            harmony_kwargs={
                'n_pcs':50
            }
        batch_correction(
            adata_hvg,
            batch_key=sample_key,
            methods='harmony',
            **harmony_kwargs
        )
        adata.obsm['X_harmony']=adata_hvg.obsm['X_harmony']
        if 'status' not in adata.uns.keys():
            adata.uns['status'] = {}
        if 'status_args' not in adata.uns.keys():
            adata.uns['status_args'] = {}

        adata.uns['status']['harmony'] = True
        adata.uns['status_args']['harmony']={
            'n_pcs':harmony_kwargs['n_pcs'],
        }
        neighbors(adata=adata,n_neighbors=15,use_rep='X_harmony',n_pcs=30)
        umap(adata)
        tsne(adata,use_rep='X_harmony')
        adata.obsm['X_umap_harmony']=adata.obsm['X_umap']
        adata.obsm['X_tsne_harmony']=adata.obsm['X_tsne']
    else:
        print('✅ Batch Correction: `Harmony` step already finished, skipping it')
    
    try:
        import scvi
        if ('X_scVI' not in adata.obsm.keys())  or ('scVI' in reforce_steps):
            print('❌ Batch Correction: `scVI` step didn\'t start, we will start it now')
            if scvi_kwargs is None:
                scvi_kwargs={
                    'n_layers':2, 
                    'n_latent':30, 
                    'gene_likelihood':"nb"
                }
            batch_correction(
                adata_hvg,
                batch_key=sample_key,
                methods='scVI',
                **scvi_kwargs
            )
            adata.obsm['X_scVI']=adata_hvg.obsm['X_scVI']
            if 'status' not in adata.uns.keys():
                adata.uns['status'] = {}
            if 'status_args' not in adata.uns.keys():
                adata.uns['status_args'] = {}

            adata.uns['status']['scVI'] = True
            adata.uns['status_args']['scVI']={
                'n_layers':scvi_kwargs['n_layers'],
                'n_latent':scvi_kwargs['n_latent'],
                'gene_likelihood':scvi_kwargs['gene_likelihood'],
            }
            neighbors(adata=adata,n_neighbors=15,use_rep='X_scVI',n_pcs=30)
            umap(adata)
            tsne(adata,use_rep='X_scVI')
            adata.obsm['X_umap_scVI']=adata.obsm['X_umap']
            adata.obsm['X_tsne_scVI']=adata.obsm['X_tsne']
        else:
            print('✅ Batch Correction: `scVI` step already finished, skipping it')
        del adata_hvg
    except:
        print('❌ scvi package not found, we will not run scVI step')
        
    

    if ('bench_best_res' not in adata.uns.keys()) or ('eval_bench' in reforce_steps):
        print('❌ Best Bench Correction Eval step didn\'t start, we will start it now')
        """
        from scib_metrics.benchmark import Benchmarker
        
        emb_keys=["X_harmony",'X_scVI']
        bm = Benchmarker(
            adata,
            batch_key=sample_key,
            label_key="phase",
            embedding_obsm_keys=emb_keys,
            pre_integrated_embedding_obsm_key="X_pca",
            n_jobs=-1,
        )
        bm.benchmark()
        bench_res = bm.get_results(min_max_scale=False)
        adata.uns['bench_res']=bench_res.loc[emb_keys]
        adata.uns['bench_best_res']=bench_res.loc[emb_keys,'Batch correction'].sort_values().index[-1]
        for col in adata.uns['bench_res']:
            adata.uns['bench_res'][col]=adata.uns['bench_res'][col].astype(float)
        import matplotlib.pyplot as plt
        bm.plot_results_table(min_max_scale=False,show=False)
        """
        if 'X_scVI' in adata.obsm.keys():
            adata.uns['bench_best_res']='X_scVI'
        else:
            adata.uns['bench_best_res']='X_harmony'
        print(f'The Best Bench Correction Method is {adata.uns["bench_best_res"]}')
        print("We can found it in `adata.uns['bench_best_res']`")
    else:
        print('✅ Best Bench Correction Eval step already finished, skipping it')

    #step 6 clusters:
    if ('best_clusters' not in adata.obs.columns)  or ('eval_clusters' in reforce_steps):
        print('❌ Best Clusters step didn\'t start, we will start it now')
        method_test=adata.uns['bench_best_res']
        print(f"Automatic clustering using sccaf")
        print(f"Dimensionality using :{method_test}")
        mde(adata,embedding_dim=2,n_neighbors=15, basis='X_mde',
                    n_pcs=30, use_rep=adata.uns['bench_best_res'],)
        neighbors(adata=adata,n_neighbors=15,use_rep=adata.uns['bench_best_res'],n_pcs=30)
        #预聚类
        print(f"Automatic clustering using leiden for preprocessed")
        sc.tl.leiden(adata, resolution=1.5, key_added = 'leiden_r1.5')
        #self.adata.obs['L1_Round0'] = self.adata.obs['leiden_r1.5']
        adata.obs['L1_result_smooth']=adata.obs['leiden_r1.5']
        #自动聚类
        for idx in range(10):
            if (len(np.unique(adata.obs['L1_result_smooth'].tolist()))>3) and idx>0:
                break
            else:
                adata.obs['L1_Round0']=adata.obs['L1_result_smooth']
                print(f"Automatic clustering using sccaf, Times: {idx}")
                from . import SCCAF_optimize_all
                SCCAF_optimize_all(min_acc=0.95, ad=adata, classifier='RF',n_jobs=4,
                                        use=adata.uns['bench_best_res'], basis ='X_mde',
                                        method='leiden',prefix='L1',plot=True)
                #平滑聚类效果
                print(f"Smoothing the effect of clustering, Times: {idx}")
                adata.obs['L1_result_smooth'] = adata.obs['L1_result'].tolist()
                
        #获取最佳聚类
        adata.obs['best_clusters']=adata.obs['L1_result_smooth'].copy()
        sc.tl.leiden(adata, resolution=1, key_added = 'leiden_clusters_L1')
        sc.tl.louvain(adata, resolution=1, key_added = 'louvain_clusters_L1')
        sc.tl.leiden(adata, resolution=0.5, key_added = 'leiden_clusters_L2')
        sc.tl.louvain(adata, resolution=0.5, key_added = 'louvain_clusters_L2')
    else:
        print('✅ Best Clusters step already finished, skipping it')

    #step 7 UMAP:
    if ('X_umap' not in adata.obsm.keys()) or ('umap' in reforce_steps):
        print('❌ UMAP step didn\'t start, we will start it now')
        umap(adata)
        adata.obsm['X_umap']=adata.obsm['X_umap']
    else:
        print('✅ UMAP step already finished, skipping it')

    #step 8 tsne:
    if ('X_tsne' not in adata.obsm.keys()) or ('tsne' in reforce_steps):
        print('❌ tSNE step didn\'t start, we will start it now')
        tsne(adata,use_rep=adata.uns['bench_best_res'])
        adata.obsm['X_tsne']=adata.obsm['X_tsne']
    else:
        print('✅ tSNE step already finished, skipping it')

    return adata

    #step 7 anno celltype automatically:




import pandas as pd
import scanpy as sc
import numpy as np

def spatial_plot(adata, info, color_by="type", size=None, mode=1, palette=None, save_path=None, figsize=None):
    """
    Plots original/mapped spatial coordinates.

    Args:
        adata: Spatial dataset to be plotted.
        info: A DataFrame containing spot/cell id and meta information.
        color_by: Which observation in adata is used to determine the color of each point in the plot.
        size: The size of each point in the plot.
        mode: Select the type of plot data. If mode=1, plot spot dataset. If mode=2, plot single-cell dataset.
        palette: Color palette for coloring the points.
        ax: Axes object to plot on.
        save_path: Path to save the plot as a PNG file.
        figsize: Tuple specifying the size of the figure in inches (width, height).

    """
    if figsize:
        plt.rcParams["figure.figsize"] = figsize

    if mode == 1:
        coor = info.copy()
        coor['id'] = coor.index.tolist()
        coor.columns = ['x', 'y', 'id']
    elif mode == 2:
        coor = info.copy()
        coor = coor[['Cell_xcoord', 'Cell_ycoord', 'cell']]
        coor.columns = ['x', 'y', 'id']

    adata = adata[adata.obs.index.isin(coor.id)]
    left_data = pd.DataFrame(adata.obs.index)
    left_data.columns = ['id']
    right_data = coor
    coordinate = pd.merge(left_data, right_data, on='id', how="left")
    coordinate.drop(['id'], axis=1, inplace=True)
    adata.obsm['spatial'] = np.array(coordinate)

    if save_path:
        sc.pl.embedding(adata, basis="spatial", color=color_by, size=size, palette=palette,show=False)
        plt.savefig(save_path)
    else:
        sc.pl.embedding(adata, basis="spatial", color=color_by, size=size, palette=palette)


import pandas as pd
import scanpy
from collections import defaultdict
import numpy as np
from scipy.optimize import nnls
import scanpy as sc
from scipy.stats import pearsonr
import matplotlib.pyplot as plt
from ..utils import pyomic_palette
import matplotlib
from typing import Union,Tuple
import anndata

def bulk2single_data_prepare(bulk_data:pd.DataFrame,
                             single_data:anndata.AnnData,celltype_key:str)->dict:
    r"""
    Prepare and align data for bulk-to-single-cell deconvolution analysis.
    
    Formats and aligns bulk RNA-seq and single-cell data by finding common genes
    and organizing metadata for downstream bulk2single analysis.

    Arguments:
        bulk_data: Bulk RNA-seq expression data with genes as rows and samples as columns
        single_data: Single-cell RNA-seq reference data as AnnData object
        celltype_key: Column name in single_data.obs containing cell type annotations

    Returns:
        dict: Dictionary containing aligned data matrices and metadata for bulk2single analysis
    """
    print("...loading data")
    input_data = {}
    
    ism=pd.DataFrame(index=single_data.obs.index)
    ism['Cell']=single_data.obs.index
    ism['Cell_type']=single_data.obs[celltype_key].values
    input_data["input_sc_meta"] = ism
    
    input_data["sc_gene"] = single_data.var.index.values.tolist()
    input_data["bulk_gene"] = bulk_data.index.values.tolist()
    
    bulk_genes=input_data["bulk_gene"]
    intersection_genes=[]
    for i in input_data["sc_gene"]:
        if i in bulk_genes:
            intersection_genes.append(i)
            
    input_data["intersect_gene"] = intersection_genes
    input_data["input_sc_data"] = single_data[:,input_data["intersect_gene"]].to_df().T
    input_data["input_bulk"] = bulk_data.loc[input_data["intersect_gene"]]
    

    return input_data

def bulk2single_plot_cellprop(generate_single_data:anndata.AnnData,
                              celltype_key:str,figsize:tuple=(4,4))->Tuple[matplotlib.figure.Figure,matplotlib.axes._axes.Axes]:
    r"""
    Plot cell-type proportions in generated single-cell data.
    
    Visualizes the distribution of cell types in data generated by bulk2single
    analysis as a bar plot with cell counts per type.

    Arguments:
        generate_single_data: Generated single-cell data from bulk2single analysis
        celltype_key: Column name in obs containing cell type annotations
        figsize: Figure dimensions as (width, height) (4, 4)

    Returns:
        matplotlib.axes.Axes: Axes object containing the cell proportion bar plot
    """
    ct_stat = pd.DataFrame(generate_single_data.obs[celltype_key].value_counts())
    generate_single_data.obs[celltype_key]=generate_single_data.obs[celltype_key].astype('category')
    key_name=list(generate_single_data.obs[celltype_key].cat.categories)
    ct_name = list(ct_stat.index)
    ct_num = list(ct_stat[celltype_key])
    if '{}_colors'.format(celltype_key) in generate_single_data.uns.keys():
        color=generate_single_data.uns['{}_colors'.format(celltype_key)]
        color_dict=dict(zip(key_name,color))
        color=[color_dict[i] for i in ct_name]
    else:
        color = pyomic_palette()
    fig, ax = plt.subplots(figsize=figsize)
    plt.bar(ct_name, ct_num, color=color)
    plt.xticks(ct_name, ct_name, rotation=90)
    plt.title("The number of cells per cell type in bulk-seq data")
    plt.xlabel("Cell type")
    plt.ylabel("Cell number")
    return ax

def bulk2single_plot_correlation(single_data:anndata.AnnData,
                                 generate_single_data:anndata.AnnData,
                                 celltype_key:str,return_table:bool=False,
                                 figsize:tuple=(6,6),cmap:str='RdBu_r')->Tuple[matplotlib.figure.Figure,matplotlib.axes._axes.Axes]:
    r"""
    Plot correlation matrix between reference and generated single-cell data.
    
    Compares expression patterns of cell types between original single-cell reference
    and bulk2single generated data using marker gene correlations.

    Arguments:
        single_data: Original single-cell reference data
        generate_single_data: Generated single-cell data from bulk2single analysis
        celltype_key: Column name in obs containing cell type annotations
        return_table: Whether to return correlation matrix instead of plot (False)
        figsize: Figure dimensions as (width, height) (6, 6)
        cmap: Colormap for correlation heatmap ('RdBu_r')

    Returns:
        matplotlib.figure.Figure: Figure object containing correlation heatmap
        matplotlib.axes.Axes: Axes object for the heatmap
    """

    # Calculate 200 marker genes of each cell type
    sc.tl.rank_genes_groups(single_data, celltype_key, method='wilcoxon')
    marker_df = pd.DataFrame(single_data.uns['rank_genes_groups']['names']).head(200)
    #marker = list(set(np.unique(np.ravel(np.array(marker_df))))&set(generate_adata.var.index.tolist()))
    marker = list(set(np.unique(np.ravel(np.array(marker_df))))&set(generate_single_data.var.index.tolist()))

    # the mean expression of 200 marker genes of input sc data
    sc_marker = single_data[:,marker].to_df()
    sc_marker[celltype_key] = single_data.obs[celltype_key]
    sc_marker_mean = sc_marker.groupby(celltype_key)[marker].mean()
    
    # the mean expression of 200 marker genes of deconvoluted bulk-seq data
    #generate_sc_meta.index = list(generate_sc_meta['Cell'])
    generate_sc_data_new = generate_single_data[:,marker].to_df()
    generate_sc_data_new[celltype_key] = generate_single_data.obs[celltype_key]
    generate_sc_marker_mean = generate_sc_data_new.groupby(celltype_key)[marker].mean()

    intersect_cell = list(set(sc_marker_mean.index).intersection(set(generate_sc_marker_mean.index)))
    generate_sc_marker_mean= generate_sc_marker_mean.loc[intersect_cell]
    sc_marker_mean= sc_marker_mean.loc[intersect_cell]

    # calculate correlation
    sc_marker_mean = sc_marker_mean.T
    generate_sc_marker_mean = generate_sc_marker_mean.T

    coeffmat = np.zeros((sc_marker_mean.shape[1], generate_sc_marker_mean.shape[1]))
    for i in range(sc_marker_mean.shape[1]):    
        for j in range(generate_sc_marker_mean.shape[1]):        
            corrtest = pearsonr(sc_marker_mean[sc_marker_mean.columns[i]], 
                                generate_sc_marker_mean[generate_sc_marker_mean.columns[j]])  
            coeffmat[i,j] = corrtest[0]
    if return_table==True:
        return coeffmat
    rf_ct = list(sc_marker_mean.columns)
    generate_ct = list(generate_sc_marker_mean.columns)

    fig, ax = plt.subplots(figsize=figsize)
    im = ax.imshow(coeffmat, cmap=cmap)
    ax.set_xticks(np.arange(len(rf_ct)))
    ax.set_xticklabels(rf_ct)
    ax.set_yticks(np.arange(len(generate_ct)))
    ax.set_yticklabels(generate_ct)
    plt.xlabel("scRNA-seq reference")
    plt.ylabel("deconvoluted bulk-seq")
    plt.setp(ax.get_xticklabels(), rotation=90, ha="right", rotation_mode="anchor")
    plt.colorbar(im)
    ax.set_title("Expression correlation")
    fig.tight_layout()
    return fig,ax



def load_data(input_bulk_path,
              input_sc_data_path,
              input_sc_meta_path,):
    r"""
    Load bulk and single-cell data from CSV files.
    
    Loads bulk RNA-seq and single-cell data from file paths and prepares
    them for bulk2single analysis by finding common genes.

    Arguments:
        input_bulk_path: Path to bulk RNA-seq CSV file (genes as rows)
        input_sc_data_path: Path to single-cell expression CSV file (genes as rows)
        input_sc_meta_path: Path to single-cell metadata CSV file

    Returns:
        dict: Dictionary containing loaded and aligned data matrices
    """
    input_sc_meta_path = input_sc_meta_path
    input_sc_data_path = input_sc_data_path
    input_bulk_path = input_bulk_path
    print("loading data......")
    input_data = {}
    # load sc_meta.csv file, containing two columns of cell name and cell type
    input_data["input_sc_meta"] = pd.read_csv(input_sc_meta_path, index_col=0)
    # load sc_data.csv file, containing gene expression of each cell
    input_sc_data = pd.read_csv(input_sc_data_path, index_col=0)
    input_data["sc_gene"] = input_sc_data.index.values.tolist()
    # load bulk.csv file, containing one column of gene expression in bulk
    input_bulk = pd.read_csv(input_bulk_path, index_col=0)
    input_data["bulk_gene"] = input_bulk.index.values.tolist()
    # filter overlapping genes.
    bulk_genes=input_data["bulk_gene"]
    intersection_genes=[]
    for i in input_data["sc_gene"]:
        if i in bulk_genes:
            intersection_genes.append(i)
    
    input_data["intersect_gene"] = intersection_genes
    input_data["input_sc_data"] = input_sc_data.loc[input_data["intersect_gene"]]
    input_data["input_bulk"] = input_bulk.loc[input_data["intersect_gene"]]
    # load st_meta.csv and st_data.csv, containing coordinates and gene expression of each spot respectively.
    #input_data["input_st_meta"] = pd.read_csv(input_st_meta_path, index_col=0)
    #input_data["input_st_data"] = pd.read_csv(input_st_data_path, index_col=0)
    print("load data done!")
    
    return input_data


def data_process(data, top_marker_num, ratio_num):
    r"""
    Process data and calculate cell-type target numbers for generation.
    
    Identifies marker genes, performs deconvolution to estimate cell-type
    proportions, and calculates target cell numbers for each type.

    Arguments:
        data: Dictionary containing aligned bulk and single-cell data
        top_marker_num: Number of top marker genes to select per cell type
        ratio_num: Multiplier for scaling target cell numbers

    Returns:
        dict: Dictionary mapping cell types to target generation numbers
    """
    # marker used
    sc = scanpy.AnnData(data["input_sc_data"].T)
    sc.obs = data["input_sc_meta"][['Cell_type']]
    scanpy.tl.rank_genes_groups(sc, 'Cell_type', method='wilcoxon')
    marker_df = pd.DataFrame(sc.uns['rank_genes_groups']['names']).head(top_marker_num)
    marker_array = np.array(marker_df)
    marker_array = np.ravel(marker_array)
    marker_array = np.unique(marker_array)
    marker = list(marker_array)
    sc_marker = data["input_sc_data"].loc[marker, :]
    bulk_marker = data["input_bulk"].loc[marker]

    #  Data processing
    breed = data["input_sc_meta"]['Cell_type']
    breed_np = breed.values
    breed_set = set(breed_np)
    id2label = sorted(list(breed_set))  # List of breed
    label2id = {label: idx for idx, label in enumerate(id2label)}  # map breed to breed-id

    cell2label = dict()  # map cell-name to breed-id
    label2cell = defaultdict(set)  # map breed-id to cell-names
    for row in data["input_sc_meta"].itertuples():
        cell_name = getattr(row, 'Cell')
        cell_type = label2id[getattr(row, 'Cell_type')]
        cell2label[cell_name] = cell_type
        label2cell[cell_type].add(cell_name)

    label_devide_data = dict()
    for label, cells in label2cell.items():
        label_devide_data[label] = sc_marker[list(cells)]

    single_cell_splitby_breed_np = {}
    for key in label_devide_data.keys():
        single_cell_splitby_breed_np[key] = label_devide_data[key].values  # [gene_num, cell_num]
        single_cell_splitby_breed_np[key] = single_cell_splitby_breed_np[key].mean(axis=1)

    max_decade = len(single_cell_splitby_breed_np.keys())
    single_cell_matrix = []

    for i in range(max_decade):
        single_cell_matrix.append(single_cell_splitby_breed_np[i].tolist())

    single_cell_matrix = np.array(single_cell_matrix)
    single_cell_matrix = np.transpose(single_cell_matrix)  # (gene_num, label_num)

    bulk_marker = bulk_marker.values  # (gene_num, 1)
    bulk_rep = bulk_marker.reshape(bulk_marker.shape[0], )

    # calculate celltype ratio in each spot by NNLS
    ratio = nnls(single_cell_matrix, bulk_rep)[0]
    ratio = ratio / sum(ratio)

    ratio_array = np.round(ratio * data["input_sc_meta"].shape[0] * ratio_num)
    ratio_list = [r for r in ratio_array]

    cell_target_num = dict(zip(id2label, ratio_list))

    return cell_target_num








class CellVote(object):

    def __init__(self, adata) -> None:
        self.adata = adata

    def popv_anno(
        self,
        ref_adata,
        ref_labels_key,
        ref_batch_key,
        query_batch_key=None,
        cl_obo_folder=None,
        save_path="tmp",
        prediction_mode="fast",
        methods=None,
        methods_kwargs=None,
    ):
        """Annotate cells using PopV.

        Parameters
        ----------
        ref_adata
            Reference :class:`anndata.AnnData` object.
        ref_labels_key
            Column in ``ref_adata.obs`` with cell-type labels.
        ref_batch_key
            Column in ``ref_adata.obs`` with batch labels.
        query_batch_key
            Batch key in query data. Defaults to ``None``.
        cl_obo_folder
            Path to ontology resources or ``None`` to disable ontology.
        save_path
            Directory to store intermediate models and predictions.
        prediction_mode
            Mode to use in PopV preprocessing. See :class:`popv.preprocessing.Process_Query`.
        methods
            Single algorithm name or list of algorithm names to run. ``None`` selects
            a default set of models based on ``prediction_mode``.
        methods_kwargs
            Dictionary of algorithm specific keyword arguments.
        """
        from ..popv.preprocessing import Process_Query
        from ..popv.annotation import annotate_data

        pq = Process_Query(
            self.adata,
            ref_adata,
            ref_labels_key=ref_labels_key,
            ref_batch_key=ref_batch_key,
            query_batch_key=query_batch_key,
            cl_obo_folder=cl_obo_folder,
            save_path_trained_models=save_path,
            prediction_mode=prediction_mode,
        ).adata

        annotate_data(
            pq,
            save_path=save_path,
            methods=methods,
            methods_kwargs=methods_kwargs,
        )
        self.adata = pq
        return pq

    def scsa_anno(self):
        """Annotate cells using the SCSA pipeline.

        This is a convenience wrapper around :class:`pySCSA` that runs the
        annotation and stores the result back into ``adata``.

        Returns
        -------
        pandas.DataFrame
            The table of annotation results produced by ``pySCSA.cell_anno``.
        """
        from ._anno import pySCSA

        scsa = pySCSA(self.adata)
        result = scsa.cell_anno()
        scsa.cell_auto_anno(self.adata)
        return result

    def gpt_anno(self):
        """Annotate cells using GPT based approach.

        The function extracts marker genes for each cluster and sends them to
        :func:`gptcelltype` for annotation. The resulting cell types are added
        to ``adata.obs['gpt_celltype']``.

        Returns
        -------
        dict
            Mapping of cluster id to predicted cell type.
        """
        from ._anno import get_celltype_marker
        from ._gptcelltype import gptcelltype

        markers = get_celltype_marker(self.adata)
        result = gptcelltype(markers)
        self.adata.obs["gpt_celltype"] = (
            self.adata.obs["leiden"].map(result).astype("category")
        )
        return result

    def gbi_anno(self):
        """Annotate clusters using GPTBioInsightor.

        The method sends cluster marker genes to
        :func:`gptbioinsightor.get_celltype` and stores the predicted cell types
        with highest score in ``adata.obs['gbi_celltype']``.

        Parameters are equivalent to those in
        :func:`gptbioinsightor.get_celltype`.

        Returns
        -------
        dict
            Score dictionary returned by GPTBioInsightor.
        """

        from gptbioinsightor import get_celltype, add_obs

        score_dic = get_celltype(self.adata)
        add_obs(self.adata, score_dic, add_key="gbi_celltype", cluster_key="leiden")
        return score_dic

    def scMulan_anno(self):
        """Annotate cells with the scMulan large language model.

        This is a thin wrapper around :mod:`omicverse.externel.scMulan` that
        performs gene symbol unification, runs the pretrained model and stores
        predictions back into ``self.adata``.

        Returns
        -------
        :class:`anndata.AnnData`
            The annotated AnnData object.
        """
        from scipy.sparse import csc_matrix
        from ..externel import scMulan
        import scanpy as sc

        # ensure CSC format for scMulan
        if not isinstance(self.adata.X, csc_matrix):
            self.adata.X = csc_matrix(self.adata.X)

        adata = scMulan.GeneSymbolUniform(
            input_adata=self.adata, output_dir="./", output_prefix="scmulan"
        )
        if adata.X.max() > 10:
            sc.pp.normalize_total(adata, target_sum=1e4)
            sc.pp.log1p(adata)

        model = scMulan.model_inference("./ckpt/ckpt_scMulan.pt", adata)
        model.cuda_count()
        model.get_cell_types_and_embds_for_adata(parallel=True, n_process=1)
        self.adata = model.adata
        return self.adata

    def vote(
        self,
        clusters_key=None,
        cluster_markers=None,
        celltype_keys=[],
        model="gpt-3.5-turbo",
        base_url=None,
        species="human",
        organization="stomach",
        provider="openai",
        result_key="CellVote_celltype",
    ):
        """
        Vote the Best celltype from scRNA-seq

        Arguments:
            clusters_key: str, the clusters key for annotation, such as leiden, louvain
            cluster_markers: dict, the markers of cluster, we can use `ov.single.get_celltype_marker` to obtain.
            celltype_keys: list, the celltype annotation columns stored in adata.obs, such as ['scsa_annotation','scMulan_anno']
            model: str, the LLM we used to identify the best matched cells in clusters.
            base_url: str, the LLM api url.
            species: str, the species of scRNA-seq,
            organization: str, the organization of scRNA-seq
            provider: str, if `base_url` is None, we can use default provider.

        Example:
        ```
        vote_obj=CellVote(adata)
        vote_obj.vote('leiden',marker_dict,
                        celltype_keys=['scsa_annotation','scMulan_anno'],
                        )
        ```
        You can found the result in adata.obs['CellVote_celltype']

        """

        cluster_celltypes = {}
        adata = self.adata
        adata.obs["best_clusters"] = adata.obs[clusters_key]
        adata.obs["best_clusters"] = adata.obs["best_clusters"].astype("category")
        for ct in adata.obs["best_clusters"].cat.categories:
            ct_li = []
            for celltype_key in celltype_keys:
                # selected the major cells as the present cells of cluster
                ct1 = (
                    adata.obs.loc[adata.obs["best_clusters"] == ct, celltype_key]
                    .value_counts()
                    .index[0]
                )
                ct_li.append(ct1)

            cluster_celltypes[ct] = ct_li

        result = get_cluster_celltype(
            cluster_celltypes,
            cluster_markers,
            species=species,
            organization=organization,
            model=model,
            base_url=base_url,
            provider=provider,
        )
        adata.obs[result_key] = (
            adata.obs["best_clusters"].map(result).astype("category")
        )
        adata.obs[result_key] = [i.capitalize() for i in adata.obs[result_key].tolist()]
        return result


def get_cluster_celltype(
    cluster_celltypes,
    cluster_markers,
    species,
    organization,
    model,
    base_url,
    provider,
    api_key=None,
):
    # from openai import OpenAI
    import os
    import numpy as np
    import pandas as pd
    import requests as requests

    if base_url is None:
        if provider == "openai":
            base_url = "https://api.openai.com/v1"
        elif provider == "kimi":
            base_url = "https://api.moonshot.cn/v1"
        elif provider == "qwen":
            base_url = "https://dashscope.aliyuncs.com/compatible-mode/v1"

    if not api_key is None:
        QWEN_API_KEY = api_key
    else:
        QWEN_API_KEY = os.getenv("AGI_API_KEY")

    # 在这里配置您在本站的API_KEY
    api_key = QWEN_API_KEY

    headers = {
        "Authorization": "Bearer " + api_key,
    }
    cluster_celltype = {}
    from tqdm import tqdm

    for cluster_id, celltypes in tqdm(cluster_celltypes.items()):
        markers = cluster_markers.get(cluster_id, [])
        question = (
            f"Given the species: {species} and organization: {organization}, "
            f"determine the most suitable cell type for cluster {cluster_id}. "
            f"The possible cell types are: {', '.join(celltypes)}. "
            f"The gene markers for this cluster are: {', '.join(markers)}. "
            f"Which cell type best represents this cluster? "
            f"Only provide the cell type name. Do not show numbers before the name. Some can be a mixture of multiple cell types."
            f"Do not provide the plural form of celltype."
        )
        # print(question)

        params = {
            "messages": [{"role": "user", "content": question}],
            # 如果需要切换模型，在这里修改
            "model": model,
        }
        response = requests.post(
            f"{base_url}/chat/completions", headers=headers, json=params, stream=False
        )
        res = response.json()
        answer = res["choices"][0]["message"]["content"].split("\n")
        # 将回答加入结果字典
        cluster_celltype[cluster_id] = answer[0].lower()

    return cluster_celltype





import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
#import ._networks as nw
from ._networks import plot_network

from collections import defaultdict


import scipy
import scanpy as sc

def plot_top_genes(adata, top_genes = 6, ncols = 2, figsize = (8,8), color_map = 'tab10', color ='attractor', attractor = None, hspace = 0.5,wspace = 0.5):
    """
    Plot the top multi-stable genes in U-S planes

    Parameters
    ----------
    adata: AnnData object
        Annotated data matrix.
    top_genes: int  
        Number of top genes to be plotted
    ncols: int
        Number of columns
    figsize: tuple
        Size of the figure
    color_map: str
        Color map for the plot
    color: str
        Color of the plot, either 'attractor' or 'membership'
    attractor: int
        Index of the attractor, if None, the average velocity will be used
    hspace: float
        Height space between subplots
    wspace: float
        Width space between subplots
    
    Returns
    -------
    None, but plots the top multi-stable genes in U-S planes
    
    """
    import scvelo as scv
    K = adata.obsm['rho'].shape[1]
    cmp = sns.color_palette(color_map, K)
    U = adata.layers['Mu']
    S = adata.layers['Ms']
    
    gene_sort = adata.var['r2_test'].sort_values(ascending=False).index.tolist()

    # calculate number of rows
    nrows = top_genes // ncols + (top_genes % ncols > 0)

    plt.figure(figsize=figsize)
    plt.subplots_adjust(hspace=hspace,wspace = wspace)
    
    for gene_id in range(top_genes):

        gene_name =  gene_sort[gene_id]
        ind_g = adata.var_names.tolist().index(gene_name)

        par = adata.uns['par'][ind_g,:]
        alpha = par[0:K]
        beta = par[K]


        ax = plt.subplot(nrows, ncols, gene_id + 1)
        
        if color == 'attractor':
            scv.pl.scatter(adata, x = S[:,ind_g],y = U[:,ind_g],color = 'attractor',show=False,alpha = 0.5,size = 50,ax=ax)
        if color == 'membership':
            scv.pl.scatter(adata, x = S[:,ind_g],y = U[:,ind_g],color = adata.obsm['rho'][:,attractor] ,show=False,size = 20,ax=ax)
        ax.axline((0, 0), slope=1/beta,color = 'k')
        ax.set_title(gene_name)
        for i in range(K):
            ax.axline((0, alpha[i]/beta), slope=0, color = cmp[i])
    return ax

def plot_genes_list(adata, genelist, ncols = 2, figsize = (8,8), color_map = 'tab10',hspace = 0.5,wspace = 0.5):
    import scvelo as scv
    K = adata.obsm['rho'].shape[1]
    cmp = sns.color_palette(color_map, K)
    U = adata.layers['Mu']
    S = adata.layers['Ms']
    
    # calculate number of rows
    nrows = len(genelist) // ncols + (len(genelist) % ncols > 0)

    plt.figure(figsize=figsize)
    plt.subplots_adjust(hspace=hspace,wspace = wspace)
    
    for gene_id in range(len(genelist)):

        gene_name =  genelist[gene_id]
        ind_g = adata.var_names.tolist().index(gene_name)

        par = adata.uns['par'][ind_g,:]
        alpha = par[0:K]
        beta = par[K]


        ax = plt.subplot(nrows, ncols, gene_id + 1)
        scv.pl.scatter(adata, x = S[:,ind_g],y = U[:,ind_g],color = 'attractor',show=False,alpha = 0.5,size = 20,ax=ax)
        ax.axline((0, 0), slope=1/beta,color = 'k')
        ax.set_title(gene_name)
        for i in range(K):
            ax.axline((0, alpha[i]/beta), slope=0, color = cmp[i])
    
            
def plot_para_hist(adata, bins = 20, log = True,figsize = (8,8)):
    gene_select = [x in adata.uns['gene_subset'] for x in adata.var_names]
    par = adata.uns['par']
    K = par.shape[1]
    fig, axs = plt.subplots(1, K, sharex=True, sharey=True, tight_layout=False, squeeze = True, figsize = figsize)

    if log:
        par = np.log10(par)
    
    for i in range(K):
        if i<K-1:
            title_name = 'alpha'+str(i)
            color = None
        else:
            title_name = 'beta'
            color = 'g'
        axs[i].hist(par[:,i],density = True, bins = bins, color = color)
        axs[i].set_xlabel('log(parameter)')
        axs[i].set_ylabel('density')
        axs[i].set_title(title_name)


def plot_sankey(vector1, vector2):
    """
    Plot a Sankey diagram. Useful to compare between annotations and attractor assignments of STT.
    
    Parameters
    ----------
    vector1: list
        A list of labels for the source nodes.
    vector2: list
        A list of labels for the target nodes.
    Returns 
    ------- 
    None
    """
    import plotly.graph_objects as go
    # Ensure both vectors have the same length.
    assert len(vector1) == len(vector2)

    label_dict = defaultdict(lambda: len(label_dict))
    
    # Generate a palette of colors.
    palette = sns.color_palette('husl', n_colors=len(set(vector1 + vector2))).as_hex()
    color_list = []

    for label in vector1 + vector2:
        label_id = label_dict[label]
        if len(color_list) <= label_id:
            color_list.append(palette[label_id % len(palette)])

    source = [label_dict[label] for label in vector1]
    target = [label_dict[label] for label in vector2]
    value = [1] * len(vector1)  # Assume each pair has a value of 1.
    
    # Color the links according to their target node.
    link_color = [color_list[target[i]] for i in range(len(target))]

    # Create the Sankey diagram.
    fig = go.Figure(data=[go.Sankey(
        node=dict(
            pad=15,
            thickness=20,
            line=dict(color="black", width=0.5),
            label=list(label_dict.keys()),
            color=color_list
        ),
        link=dict(
            source=source,
            target=target,
            value=value,
            color=link_color  # Color the links.
        )
    )])

    fig.update_layout(title_text="Sankey Diagram", font_size=20)
    fig.show()  

                
def compute_tensor_similarity(adata, adata_aggr, pathway1, pathway2, state = 'spliced', attractor = None):
    """
    Compute the similarity between two pathways based on the tensor graph
    
    Parameters
    ----------
    adata: AnnData object
    adata_aggr: AnnData object
    pathway1: list
        List of genes in the first pathway
    pathway2: list
        List of genes in the second pathway
    state: str
        State of the tensor graph, either 'spliced', 'unspliced', or 'joint'
    attractor: int
        Index of the attractor, if None, the average velocity will be used
    
    Returns 
    -------
    float, the correlation coefficient between the two pathways       
    """
    import scvelo as scv
    if attractor == None:
        velo =  adata.obsm['tensor_v_aver'].copy()
    else:
        velo = adata.obsm['tensor_v'][:,:,:,attractor].copy()
    
    if state == 'spliced':
        vkey = 'vs'
        xkey = 'Ms'
    if state == 'unspliced':
        vkey = 'vu'
        xkey = 'Mu'
    if state == 'joint':
        print("check that the input includes aggregated object") # some problem needs fixed
        adata_aggr.layers['vj'] = np.concatenate((velo[:,:,0],velo[:,:,1]),axis = 1)
        vkey = 'vj'
        xkey = 'Ms'

    scv.tl.velocity_graph(adata, vkey = vkey, xkey = xkey, gene_subset = pathway1,n_jobs = -1)
    tpm1 = adata.uns[vkey+'_graph'].toarray()
    scv.tl.velocity_graph(adata, vkey = vkey, xkey = vkey, gene_subset = pathway2,n_jobs = -1)
    tpm2 = adata.uns[vkey+'_graph'].toarray()
    return np.corrcoef(tpm1.reshape(-1),tpm2.reshape(-1))[0,1]

def plot_landscape(sc_object,show_colorbar = False, dim = 2, size_point = 3, 
                   alpha_land = 0.5, alpha_point = 0.5,  color_palette_name = 'Set1',
                     contour_levels = 15, elev=10, azim = 4):
    """
    Plot the landscape of the attractor landscape
    
    Parameters
    ----------
    sc_object : AnnData object
        Annotated data matrix.
    show_colorbar : bool
        Whether to show the colorbar
    dim : int
        Dimension of the plot
    size_point : float
        Size of the points
    alpha_land : float
        Transparency of the landscape
    alpha_point : float 
        Transparency of the points
    color_palette_name : str
        Name of the color palette
    contour_levels : int
        Number of contour levels
    elev : int  
        Elevation of the 3D plot
    azim : int  
        Azimuth of the 3D plot
    
    Returns 
    -------
    None            
    """
    land_value = sc_object.uns['land_out']['land_value']
    xv = sc_object.uns['land_out']['grid_x']
    yv = sc_object.uns['land_out']['grid_y']
    trans_coord = sc_object.uns['land_out']['trans_coord'] 
    land_cell = sc_object.obs['land_cell']
    
    K = sc_object.obsm['rho'].shape[1]
    labels =sc_object.obs['attractor'].astype(int)
    
    color_palette = sns.color_palette(color_palette_name, K)
    cluster_colors = [color_palette[x] for x in labels]
    
    if dim == 2:
        plt.contourf(xv, yv, land_value, levels=contour_levels, cmap="Greys_r",zorder=-100, alpha = alpha_land)
        plt.scatter(*trans_coord.T, s=size_point, linewidth=0, c=cluster_colors, alpha=alpha_point)

    else:
        ax = plt.axes(projection='3d')
        ax.scatter(*trans_coord.T,land_cell,s=size_point, linewidth=0, c=cluster_colors, alpha=alpha_point)
        ax.plot_surface(xv, yv,land_value,rstride=1, cstride=1, linewidth=0, antialiased=True,cmap="Greys_r", alpha = alpha_land, vmin = 0, vmax=np.nanmax(land_value), shade = True)
        ax.grid(False)
        ax.axis('off')
        ax.view_init(elev=elev, azim=azim)
        
    if show_colorbar:
        plt.colorbar()
        

def infer_lineage(sc_object,si=0,sf=1,method = 'MPFT',
                  flux_fraction = 0.9, size_state = 0.1, 
                  size_point = 3, alpha_land = 0.5, alpha_point = 0.5, 
                  size_text=20, show_colorbar = False, 
                  color_palette_name = 'Set1', contour_levels = 15,ax = None):
    """
    Infer the lineage among the multi-stable attractors based on most probable flux tree or path
    
    Parameters
    ----------
    sc_object : AnnData object
        Annotated data matrix.
    si : int or list
        Initial state (attractor index number) of specified transition path, specified when method = 'MPPT'
    sf : int or list
        Final state (attractor index number) , specified when method = 'MPPT'
    method : str
        Method to infer the lineage, either 'MPFT'(maxium probability flow tree, global) or 'MPPT'(most probable path tree, local)
    flux_fraction : float
        Fraction of the total flux to be considered
    size_state : float  
        Size of the state
    size_point : float
        Size of the points
    alpha_land : float
        Transparency of the landscape
    alpha_point : float 
        Transparency of the points
    size_text : float
        Size of the text
    show_colorbar : bool
        Whether to show the colorbar
    color_palette_name : str
        Name of the color palette
    contour_levels : int
        Number of contour levels
    
    Returns 
    -------
    None
    """
    import pyemma.msm as msm
    K = sc_object.obsm['rho'].shape[1]
    centers = sc_object.uns['land_out']['cluster_centers']

    

    P_hat = sc_object.uns['da_out']['P_hat']
    M = msm.markov_model(P_hat)
    mu_hat = M.pi

    if ax == None:
        fig,ax= plt.subplots(1,1,figsize=(4,4))
    
    if method == 'MPFT':
        Flux_cg = np.diag(mu_hat.reshape(-1)).dot(P_hat)
        max_flux_tree = scipy.sparse.csgraph.minimum_spanning_tree(-Flux_cg)
        max_flux_tree = -max_flux_tree.toarray()
        #for i in range(K):
        #    for j in range(i+1,K):   
        #        max_flux_tree[i,j]= max(max_flux_tree[i,j],max_flux_tree[j,i])
        #        max_flux_tree[j,i] = max_flux_tree[i,j]
        
        plot_network(max_flux_tree, pos=centers, state_scale=size_state, state_sizes=mu_hat, arrow_scale=2.0,arrow_labels= None, arrow_curvature = 0.2, ax=ax,max_width=1000, max_height=1000)
        plot_landscape(sc_object, show_colorbar = show_colorbar, size_point = size_point, alpha_land = alpha_land, alpha_point = alpha_point,  color_palette_name = color_palette_name)
        plt.axis('off')  
        

        
    if method == 'MPPT':
        
        #state_reorder = np.array(range(K))
        #state_reorder[0] = si
        #state_reorder[-1] = sf
        #state_reorder[sf+1:-1]=state_reorder[sf+1:-1]+1
        #state_reorder[1:si]=state_reorder[1:si]-1
        if isinstance(si,int):
            si = list(map(int, str(si)))
        
        if isinstance(sf,int):
            sf = list(map(int, str(sf)))
        import pyemma.msm as msm
        
        tpt = msm.tpt(M, si, sf)
        Fsub = tpt.major_flux(fraction=flux_fraction)
        Fsubpercent = 100.0 * Fsub / tpt.total_flux
        
    
        plot_landscape(sc_object, show_colorbar = show_colorbar, size_point = size_point, alpha_land = alpha_land, alpha_point = alpha_point,  color_palette_name = color_palette_name, contour_levels = contour_levels)
        plot_network(Fsubpercent, state_scale=size_state*mu_hat,pos=centers, arrow_label_format="%3.1f",arrow_label_size = size_text,ax=plt.gca(), max_width=1000, max_height=1000)
        plt.axis('off')   

    return ax 

def plot_tensor_heatmap(adata, attractor = 'all', component = 'spliced', top_genes = 50,ax=None):
    """
    Plot the heatmap of the transition tensor

    Parameters
    ----------
    adata: AnnData object
        Annotated data matrix.
    attractor: int
        Index of the attractor, if None, the average velocity will be used
    component: str
        Component of the tensor, either 'spliced' or 'unspliced'
    top_genes: int
        Number of top genes to be plotted
    
    Returns
    -------
    None
    """

    gene_sort = adata.var['r2_test'].sort_values(ascending=False).index.tolist()
    if component == 'unspliced':
        component_ind = 0
    else:
        component_ind = 1

    if attractor == 'all':
        adata.layers['velo_plot'] = adata.obsm['tensor_v_aver'][:,:,component_ind]
    else:
        adata.layers['velo_plot'] = adata.obsm['tensor_v'][:,:,component_ind,attractor]
    sc.pl.heatmap(adata, gene_sort[0:top_genes], groupby='attractor', layer = 'velo_plot',standard_scale = 'var',cmap='RdBu_r',
                 )
    plt.suptitle('Tensor of ' + component + ', Attractor '+str(attractor))
    return ax
    


import numpy as np
import pandas as pd
import scanpy as sc
from typing import Union, Sequence

def get_top_nmf_genes(adata: sc.AnnData,
                      gems: Union[str, Sequence[str]],
                      n_genes: int,
                      gene_type: str = 'all',
                      model_organism: str = 'human'):
    import pkg_resources
    tf_path=pkg_resources.resource_filename("omicverse", 'data_files/TF/allTFs_' + model_organism + '.txt')
    gene_types = ['all', 'tf']

    if gene_type not in gene_types:
        raise ValueError ("Invalid gene type. Please select one of: %s" % gene_types)
    
    model_organisms = ['human', 'mouse']

    if model_organism not in model_organisms:
        raise ValueError ("Invalid model organism. Please select one of: %s" % model_organisms)
    
    # Load the TFs if we need to
    if gene_type == 'tf':
        tfs_list = pd.read_csv(tf_path, header=None)[0].tolist()

    n_gems = adata.uns['nmf_info']['n_gems']
    nmf_vars = np.array(adata.uns['nmf_info']['vars'], dtype=object)
    nmf_loadings = adata.uns['nmf_info']['loadings']

    all_gems = ['GEM-' + str(i + 1) for i in range(n_gems)]

    top_genes_in_gems = []
    gem_labels = []
    gem_weights = []

    for gem in gems:
         
        gem_index = all_gems.index(gem)

        # Get the loadings corresponding tot his gem
        gem_loadings = nmf_loadings[gem_index, :]

        # Sort the genes in their order by loading
        sorted_gem_loadings = gem_loadings[np.argsort(-gem_loadings)]
        ordered_genes = nmf_vars[np.argsort(-gem_loadings)]

        if gene_type == 'all':
              
            for i in range(n_genes):

                top_genes_in_gems.append(ordered_genes[i])
                gem_labels.append(gem)
                gem_weights.append(sorted_gem_loadings[i])

        else: # We only take TFs from each GEM
         
            ordered_tfs = [gene for gene in ordered_genes if gene in tfs_list]
            sorted_tf_loadings = [sorted_gem_loadings[i] for i, gene in enumerate(ordered_genes) if gene in tfs_list]

            for i in range(n_genes):

                top_genes_in_gems.append(ordered_tfs[i])
                gem_labels.append(gem)
                gem_weights.append(sorted_tf_loadings[i])

    top_nmf_genes_df = pd.DataFrame(data={'Gene': top_genes_in_gems,
                                          'GEM': gem_labels,
                                          'Weight': gem_weights})
    return top_nmf_genes_df

def get_top_pyliger_genes(adata: sc.AnnData,
                        gems: Union[str, Sequence[str]],
                        n_genes: int,
                        gene_type: str = 'all',
                        model_organism: str = 'human'):
    
    import pkg_resources
    tf_path=pkg_resources.resource_filename("omicverse", 'data_files/TF/allTFs_' + model_organism + '.txt')

    gene_types = ['all', 'tf']

    if gene_type not in gene_types:
        raise ValueError ("Invalid gene type. Please select one of: %s" % gene_types)
    
    model_organisms = ['human', 'mouse']

    if model_organism not in model_organisms:
        raise ValueError ("Invalid model organism. Please select one of: %s" % model_organisms)

    # Load the TFs if we need to
    if gene_type == 'tf':
        tfs_list = pd.read_csv(tf_path, header=None)[0].tolist()

    pyliger_conds = [key for key in adata.uns['pyliger_info'].keys() if key not in ['n_vars', 'vars','n_gems']]

    n_gems = adata.uns['pyliger_info']['n_gems']
    pyliger_vars = np.array(adata.uns['pyliger_info']['vars'], dtype=object)

    all_gems = ['GEM-' + str(i + 1) for i in range(n_gems)]

    top_genes_in_gems = []
    gem_labels = []
    gem_weights = []

    for gem in gems:
        
        gem_index = all_gems.index(gem)

        all_gem_loadings = {cond: adata.uns['pyliger_info'][cond]['W'][:, gem_index] \
                                    + adata.uns['pyliger_info'][cond]['V'][:, gem_index] for cond in pyliger_conds}

        stacked_gem_loadings = np.hstack([all_gem_loadings[cond] for cond in all_gem_loadings])
        stacked_gem_genes = np.hstack([pyliger_vars for cond in all_gem_loadings])     
        sorted_gem_loadings = stacked_gem_loadings[np.argsort(-stacked_gem_loadings)]
        init_top_gem_genes = stacked_gem_genes[np.argsort(-stacked_gem_loadings)]
        
        top_genes = []

        for i, gene in enumerate(init_top_gem_genes):

            if gene_type == 'tf':

                if (gene in tfs_list)&(gene not in top_genes):
                    
                    top_genes.append(gene) # This tracks repeats, given that we're stacking the list of genes 
                    top_genes_in_gems.append(gene)
                    gem_labels.append(gem)
                    gem_weights.append(sorted_gem_loadings[i])
                    
                    if len(top_genes) == n_genes:
                        break
            else:

                if gene not in top_genes:

                    top_genes.append(gene) # This tracks repeated names, given that we're stacking the list of genes
                    top_genes_in_gems.append(gene)
                    gem_labels.append(gem)
                    gem_weights.append(sorted_gem_loadings[i])
                    
                if len(top_genes) == n_genes:
                    break

        

    top_pyliger_genes_df = pd.DataFrame(data={'Gene': top_genes_in_gems,
                                          'GEM': gem_labels,
                                          'Weight': gem_weights})
    return top_pyliger_genes_df

def get_top_nsf_genes(adata: sc.AnnData,
                      gems: Union[str, Sequence[str]],
                      n_genes: int,
                      gene_type: str = 'all',
                      model_organism: str = 'human'):
    import pkg_resources
    tf_path=pkg_resources.resource_filename("omicverse", 'data_files/TF/allTFs_' + model_organism + '.txt')
    
    gene_types = ['all', 'tf']

    if gene_type not in gene_types:
        raise ValueError ("Invalid gene type. Please select one of: %s" % gene_types)
    
    model_organisms = ['human', 'mouse']

    if model_organism not in model_organisms:
        raise ValueError ("Invalid model organism. Please select one of: %s" % model_organisms)
    
    # Load the TFs if we need to
    if gene_type == 'tf':
        tfs_list = pd.read_csv(tf_path, header=None)[0].tolist()

    n_gems = adata.uns['nsf_info']['n_gems']
    nsf_vars = np.array(adata.uns['nsf_info']['vars'], dtype=object)
    nsf_loadings = adata.uns['nsf_info']['loadings'].T

    all_gems = ['GEM-' + str(i + 1) for i in range(n_gems)]

    top_genes_in_gems = []
    gem_labels = []
    gem_weights = []

    for gem in gems:
         
        gem_index = all_gems.index(gem)

        # Get the loadings corresponding tot his gem
        gem_loadings = nsf_loadings[gem_index, :]

        # Sort the genes in their order by loading
        sorted_gem_loadings = gem_loadings[np.argsort(-gem_loadings)]
        ordered_genes = nsf_vars[np.argsort(-gem_loadings)]

        if gene_type == 'all':
              
            for i in range(n_genes):

                top_genes_in_gems.append(ordered_genes[i])
                gem_labels.append(gem)
                gem_weights.append(sorted_gem_loadings[i])

        else: # We only take TFs from each GEM
         
            ordered_tfs = [gene for gene in ordered_genes if gene in tfs_list]
            sorted_tf_loadings = [sorted_gem_loadings[i] for i, gene in enumerate(ordered_genes) if gene in tfs_list]

            for i in range(n_genes):

                top_genes_in_gems.append(ordered_tfs[i])
                gem_labels.append(gem)
                gem_weights.append(sorted_tf_loadings[i])

    top_nsf_genes_df = pd.DataFrame(data={'Gene': top_genes_in_gems,
                                          'GEM': gem_labels,
                                          'Weight': gem_weights})
    return top_nsf_genes_df

def get_top_gem_genes(adata: sc.AnnData,
                      gems: Union[str, Sequence[str]],
                      n_genes: int,
                      method: str = 'pyliger',
                      gene_type: str = 'all',
                      model_organism: str = 'human'):
    
    # Perform a bunch of checks that we're specifying the right options
    methods = ['nmf', 'pyliger', 'nsf']
    if method not in methods:
                raise ValueError ("Invalid method. Please select one of: %s" % methods)
    
    gene_types = ['all', 'tf']

    if gene_type not in gene_types:
        raise ValueError ("Invalid gene type. Please select one of: %s" % gene_types)
    
    model_organisms = ['human', 'mouse']

    if model_organism not in model_organisms:
        raise ValueError ("Invalid model organism. Please select one of: %s" % model_organisms)
    
    if method == 'nmf':
          
          return get_top_nmf_genes(adata, gems, n_genes, gene_type, model_organism)
    
    elif method == 'pyliger':
          
        return get_top_pyliger_genes(adata, gems, n_genes, gene_type, model_organism)
    
    else: # Should be NSF
    
        return get_top_nsf_genes(adata, gems, n_genes, gene_type, model_organism)



# __init__ file
from ._cot import cot_dense
from ._cot import cot_row_dense
from ._cot import cot_col_dense
from ._cot import cot_blk_dense
from ._cot import cot_combine_sparse
from ._cot import cot_sparse
from ._cot import cot_row_sparse
from ._cot import cot_col_sparse
from ._cot import cot_blk_sparse
from ._usot import usot
from ._usot import uot
from ._unot import unot
from ._unot_torch import unot_torch, unot_sinkhorn_l1_dense_torch, unot_sinkhorn_l1_sparse_torch, unot_momentum_l1_dense_torch, unot_sinkhorn_l2_dense_torch, to_numpy, to_torch


'''
Functions for computing evaluation metrics.
Some codes used for GRN evaluation are referenced from https://github.com/murali-group/Beeline.
'''
import pandas as pd
import numpy as np
from sklearn.metrics import precision_recall_curve, roc_curve, auc
from itertools import product, permutations
from scipy.stats import hypergeom


def EarlyPrec(trueEdgesDF: pd.DataFrame, predEdgesDF: pd.DataFrame,
              weight_key: str = 'weights_combined', TFEdges: bool = True):
    """
        Computes early precision for a given set of predictions in the form of a DataFrame.
        The early precision is defined as the fraction of true positives in the top-k edges,
        where k is the number of edges in the ground truth network (excluding self loops).

    :param trueEdgesDF:   A pandas dataframe containing the true edges.
    :type trueEdgesDF: DataFrame

    :param predEdgesDF:   A pandas dataframe containing the edges and their weights from
        the predicted network. Use param `weight_key` to assign the column name of edge weights.
        Higher the weight, higher the edge confidence.
    :type predEdgesDF: DataFrame

    :param weight_key:   A str represents the column name containing weights in predEdgeDF.
    :type weight_key: str

    :param TFEdges:   A flag to indicate whether to consider only edges going out of TFs (TFEdges = True)
        or not (TFEdges = False) from evaluation.
    :type TFEdges: bool

    :returns:
        - Eprec: Early precision value
        - Erec: Early recall value
        - EPR: Early precision ratio
    """

    print("Calculating the EPR(early prediction rate)...")

    # Remove self-loops
    trueEdgesDF = trueEdgesDF.loc[(trueEdgesDF['Gene1'] != trueEdgesDF['Gene2'])]
    if 'Score' in trueEdgesDF.columns:
        trueEdgesDF = trueEdgesDF.sort_values('Score', ascending=False)
    trueEdgesDF = trueEdgesDF.drop_duplicates(keep='first', inplace=False).copy()
    trueEdgesDF.reset_index(drop=True, inplace=True)

    predEdgesDF = predEdgesDF.loc[(predEdgesDF['Gene1'] != predEdgesDF['Gene2'])]
    if weight_key in predEdgesDF.columns:
        predEdgesDF = predEdgesDF.sort_values(weight_key, ascending=False)
    predEdgesDF = predEdgesDF.drop_duplicates(keep='first', inplace=False).copy()
    predEdgesDF.reset_index(drop=True, inplace=True)

    uniqueNodes = np.unique(trueEdgesDF.loc[:, ['Gene1', 'Gene2']])
    if TFEdges:
        # Consider only edges going out of source genes
        print("  Consider only edges going out of source genes")

        # Get a list of all possible TF to gene interactions
        possibleEdges_TF = set(product(set(trueEdgesDF.Gene1), set(uniqueNodes)))

        # Get a list of all possible interactions 
        possibleEdges_noSelf = set(permutations(uniqueNodes, r=2))

        # Find intersection of above lists to ignore self edges
        # TODO: is there a better way of doing this?
        possibleEdges = possibleEdges_TF.intersection(possibleEdges_noSelf)
        possibleEdges = pd.DataFrame(possibleEdges, columns=['Gene1', 'Gene2'], dtype=str)

        # possibleEdgesDict = {'|'.join(p): 0 for p in possibleEdges}
        possibleEdgesDict = possibleEdges['Gene1'] + "|" + possibleEdges['Gene2']

        trueEdges = trueEdgesDF['Gene1'].astype(str) + "|" + trueEdgesDF['Gene2'].astype(str)
        trueEdges = trueEdges[trueEdges.isin(possibleEdgesDict)]
        print("  {} TF Edges in ground-truth".format(len(trueEdges)))
        numEdges = len(trueEdges)

        predEdgesDF['Edges'] = predEdgesDF['Gene1'].astype(str) + "|" + predEdgesDF['Gene2'].astype(str)
        # limit the predicted edges to the genes that are in the ground truth
        predEdgesDF = predEdgesDF[predEdgesDF['Edges'].isin(possibleEdgesDict)]
        print("  {} Predicted TF edges are considered".format(len(predEdgesDF)))

        M = len(set(trueEdgesDF.Gene1)) * (len(uniqueNodes) - 1)

    else:
        trueEdges = trueEdgesDF['Gene1'].astype(str) + "|" + trueEdgesDF['Gene2'].astype(str)
        trueEdges = set(trueEdges.values)
        numEdges = len(trueEdges)
        print("  {} edges in ground-truth".format(len(trueEdges)))

        M = len(uniqueNodes) * (len(uniqueNodes) - 1)

    if not predEdgesDF.shape[0] == 0:
        # Use num True edges or the number of
        # edges in the dataframe, which ever is lower
        maxk = min(predEdgesDF.shape[0], numEdges)
        edgeWeightTopk = predEdgesDF.iloc[maxk - 1][weight_key]

        nonZeroMin = np.nanmin(predEdgesDF[weight_key].values)
        bestVal = max(nonZeroMin, edgeWeightTopk)

        newDF = predEdgesDF.loc[(predEdgesDF[weight_key] >= bestVal)]
        predEdges = set(newDF['Gene1'].astype(str) + "|" + newDF['Gene2'].astype(str))
        print("  {} Top-k edges selected".format(len(predEdges)))
    else:
        predEdges = set([])

    if len(predEdges) != 0:
        intersectionSet = predEdges.intersection(trueEdges)
        print("  {} true-positive edges".format(len(intersectionSet)))
        Eprec = len(intersectionSet) / len(predEdges)
        Erec = len(intersectionSet) / len(trueEdges)
    else:
        Eprec = 0
        Erec = 0

    random_EP = len(trueEdges) / M
    EPR = Erec / random_EP
    return Eprec, Erec, EPR


def computeScores(trueEdgesDF: pd.DataFrame, predEdgesDF: pd.DataFrame,
                  weight_key: str = 'weights_combined', selfEdges: bool = True):
    """
        Computes precision-recall and ROC curves using scikit-learn
        for a given set of predictions in the form of a DataFrame.

    :param trueEdgesDF:   A pandas dataframe containing the true edges.
    :type trueEdgesDF: DataFrame

    :param predEdgesDF:   A pandas dataframe containing the edges and their weights from
        the predicted network. Use param `weight_key` to assign the column name of edge weights.
        Higher the weight, higher the edge confidence.
    :type predEdgesDF: DataFrame

    :param weight_key:   A str represents the column name containing weights in predEdgeDF.
    :type weight_key: str

    :param selfEdges:   A flag to indicate whether to include self-edges (selfEdges = True)
        or exclude self-edges (selfEdges = False) from evaluation.
    :type selfEdges: bool

    :returns:
        - prec: A list of precision values (for PR plot)
        - recall: A list of precision values (for PR plot)
        - fpr: A list of false positive rates (for ROC plot)
        - tpr: A list of true positive rates (for ROC plot)
        - AUPRC: Area under the precision-recall curve
        - AUROC: Area under the ROC curve
    """

    from rpy2.robjects.packages import importr
    from rpy2.robjects import FloatVector
    print("Calculating the AUPRC and AUROC...")

    if 'Score' in trueEdgesDF.columns:
        trueEdgesDF = trueEdgesDF.sort_values('Score', ascending=False)
    trueEdgesDF = trueEdgesDF.drop_duplicates(keep='first', inplace=False).copy()
    trueEdgesDF.reset_index(drop=True, inplace=True)

    if weight_key in predEdgesDF.columns:
        predEdgesDF = predEdgesDF.sort_values(weight_key, ascending=False)
    predEdgesDF = predEdgesDF.drop_duplicates(keep='first', inplace=False).copy()
    predEdgesDF.reset_index(drop=True, inplace=True)

    # Initialize dictionaries with all
    # possible edges
    if selfEdges:
        possibleEdges = list(product(np.unique(trueEdgesDF.loc[:, ['Gene1', 'Gene2']]),
                                     repeat=2))
    else:
        possibleEdges = list(permutations(np.unique(trueEdgesDF.loc[:, ['Gene1', 'Gene2']]),
                                          r=2))
        trueEdgesDF = trueEdgesDF.loc[(trueEdgesDF['Gene1'] != trueEdgesDF['Gene2'])]
        predEdgesDF = predEdgesDF.loc[(predEdgesDF['Gene1'] != predEdgesDF['Gene2'])]

    TrueEdgeDict = pd.DataFrame({'|'.join(p): 0 for p in possibleEdges}, index=['label']).T
    PredEdgeDict = pd.DataFrame({'|'.join(p): 0 for p in possibleEdges}, index=['label']).T

    # Compute TrueEdgeDict Dictionary
    # 1 if edge is present in the ground-truth
    # 0 if edge is not present in the ground-truth
    TrueEdgeDict.loc[np.array(trueEdgesDF['Gene1'] + "|" + trueEdgesDF['Gene2']), 'label'] = 1
    PredEdgeDict.loc[np.array(predEdgesDF['Gene1'] + "|" + predEdgesDF['Gene2']), 'label'] = np.abs(
        predEdgesDF[weight_key].values)

    # Combine into one dataframe
    # to pass it to sklearn
    outDF = pd.DataFrame([TrueEdgeDict['label'].values, PredEdgeDict['label'].values]).T
    outDF.columns = ['TrueEdges', 'PredEdges']
    prroc = importr('PRROC')
    prCurve = prroc.pr_curve(scores_class0=FloatVector(list(outDF['PredEdges'].values)),
                             weights_class0=FloatVector(list(outDF['TrueEdges'].values)))

    fpr, tpr, thresholds = roc_curve(y_true=outDF['TrueEdges'],
                                     y_score=outDF['PredEdges'],
                                     drop_intermediate=True, pos_label=1)

    prec, recall, thresholds = precision_recall_curve(y_true=outDF['TrueEdges'],
                                                      probas_pred=outDF['PredEdges'], pos_label=1)

    return prec, recall, fpr, tpr, prCurve[2][0], auc(fpr, tpr)


def gene_TopK_precision(trueGenesDF: pd.DataFrame, predGenesDF: pd.DataFrame,
                        weight_key: str, K: int):
    """
        Computes the precision of the top K predicted genes.

    :param trueGenesDF:   A pandas dataframe containing the ground-truth gene set.
        The order of genes must be sorted.
    :type trueGenesDF: DataFrame

    :param predGenesDF:   A pandas dataframe containing the predicted gene set with their weights.
        Higher the weight, more important.
    :type predGenesDF: DataFrame

    :param weight_key:   A str represents the column name containing weights in predGenesDF.
    :type weight_key: str

    :param K:   The number of top genes to be considered for precision.
    :type K: int

    :returns:
        - precision: the precision
        - recall: the recall
        - P_value: the p-value of hypergeometric distribution test
    """

    trueGenesDF = trueGenesDF.iloc[:, 0].values
    predGenesDF = predGenesDF.sort_values(weight_key, ascending=False)
    predGenesK = set(predGenesDF.iloc[:K, 'Gene'].values)

    TP = len(predGenesK.intersection(set(trueGenesDF)))
    precision = TP / len(predGenesK)
    recall = TP / len(set(trueGenesDF))
    [k, M, n, N] = [TP, 15000, len(set(trueGenesDF)), len(predGenesK)]
    P_value = hypergeom.sf(k, M, n, N)  # P-value of hypergeometric distribution test. k, M, n, N,

    return precision, recall, P_value



r"""
Preprocessing utilities for single-cell and bulk omics data.

This module provides essential preprocessing functions for quality control,
normalization, dimensionality reduction, and neighborhood analysis of omics data.
It supports both CPU and GPU computation for large-scale datasets.

Core preprocessing functions:
    preprocess: Complete preprocessing pipeline
    normalize_pearson_residuals: Pearson residuals normalization
    highly_variable_genes: Identify highly variable genes
    scale: Z-score scaling of expression data
    regress: Regress out unwanted variation
    
Dimensionality reduction:
    pca: Principal component analysis
    mde: Minimum distortion embedding
    tsne: t-SNE embedding
    umap: UMAP embedding
    
Neighborhood analysis:
    neighbors: Compute neighborhood graph
    leiden: Leiden clustering
    louvain: Louvain clustering
    
Quality control:
    quantity_control: Comprehensive QC metrics
    filter_cells: Remove low-quality cells
    filter_genes: Remove low-quality genes
    
Utility functions:
    anndata_to_GPU: Transfer data to GPU
    anndata_to_CPU: Transfer data to CPU
    recover_counts: Recover original count data
    score_genes_cell_cycle: Cell cycle scoring

Examples:
    >>> import omicverse as ov
    >>> # Basic preprocessing pipeline
    >>> ov.pp.preprocess(adata, mode='shiftlog|pearson', n_HVGs=2000)
    >>> ov.pp.scale(adata)
    >>> ov.pp.pca(adata)
    >>> 
    >>> # Quality control
    >>> ov.pp.qc(adata, tresh={'mito_perc': 20, 'nUMIs': 1000})
    >>> ov.pp.filter_cells(adata, min_genes=200)
    >>> ov.pp.filter_genes(adata, min_cells=3)
    
    >>> # Neighborhood analysis
    >>> ov.pp.neighbors(adata, n_neighbors=15, n_pcs=50)
    >>> ov.pp.leiden(adata, resolution=0.5)
    >>> ov.pp.umap(adata)
"""

from ._preprocess import (identify_robust_genes,
                          select_hvf_pegasus,
                          highly_variable_features,
                          remove_cc_genes,
                          preprocess,
                          normalize_pearson_residuals,
                          highly_variable_genes,
                          scale,
                          regress,
                          regress_and_scale,
                          neighbors,
                          pca,score_genes_cell_cycle,
                          leiden,umap,louvain,anndata_to_GPU,anndata_to_CPU,mde,tsne)

from ._qc import quantity_control,qc,filter_cells,filter_genes
from ._recover import recover_counts,binary_search

__all__ = [
    # Core preprocessing
    'identify_robust_genes',
    'select_hvf_pegasus',
    'highly_variable_features',
    'remove_cc_genes',
    'preprocess',
    'normalize_pearson_residuals',
    'highly_variable_genes',
    'scale',
    'regress',
    'regress_and_scale',
    
    # Dimensionality reduction
    'pca',
    'mde',
    'tsne',
    'umap',
    
    # Neighborhood and clustering
    'neighbors',
    'leiden',
    'louvain',
    
    # Quality control
    'quantity_control',
    'qc',
    'filter_cells',
    'filter_genes',
    
    # Utility functions
    'score_genes_cell_cycle',
    'anndata_to_GPU',
    'anndata_to_CPU',
    'recover_counts',
    'binary_search',
]


r"""
Single-cell omics analysis utilities.

This module provides comprehensive tools for single-cell RNA-seq analysis including:
- Cell type annotation and automated identification
- Trajectory inference and pseudotime analysis
- Gene regulatory network analysis
- Cell-cell communication analysis
- Multi-modal integration (RNA + ATAC)
- Drug response prediction
- Pathway enrichment and functional analysis

Key classes:
    pyMOFA: Multi-Omics Factor Analysis
    pyVIA: Velocity and pseudotime analysis 
    pySIMBA: Single-cell integration and batch alignment
    pyTOSICA: Trajectory inference and cell fate analysis
    pyCEFCON: Cell-cell communication analysis
    MetaCell: Metacell construction and analysis
    pyDEG: Single-cell differential expression

Annotation tools:
    pySCSA: Automated cell type annotation
    MetaTiME: Tumor microenvironment annotation
    gptcelltype: GPT-based cell type identification
    CellVote: Consensus cell type annotation

Trajectory analysis:
    TrajInfer: Trajectory inference framework
    scLTNN: Lineage tracing with neural networks
    cytotrace2: Developmental potential scoring

Examples:
    >>> import omicverse as ov
    >>> # Automated cell annotation
    >>> ov.single.pySCSA(adata, tissue='lung', species='human')
    >>> 
    >>> # Trajectory analysis
    >>> via = ov.single.pyVIA(adata)
    >>> via.run_via()
    >>> 
    >>> # Multi-omics integration
    >>> mofa = ov.single.pyMOFA(data_dict)
    >>> mofa.build_mofa()
"""

# Heavy functionality lives in submodules and is imported lazily.

# Core gene selection and differential expression
from ._cosg import cosg
from ._anno import pySCSA,MetaTiME,scanpy_lazy,scanpy_cellanno_from_dict,get_celltype_marker
from ._nocd import scnocd
from ._mofa import pyMOFAART,pyMOFA,GLUE_pair,factor_exact,factor_correlation,get_weights,glue_pair,get_r2_from_hdf5
from ._scdrug import autoResolution,writeGEP,Drug_Response
from ._cpdb import (cpdb_network_cal,cpdb_plot_network,
                    cpdb_plot_interaction,
                    cpdb_interaction_filtered,
                    cpdb_submeans_exacted,cpdb_exact_target,cpdb_exact_source,cellphonedb_v5,run_cellphonedb_v5)
from ._scgsea import (geneset_aucell,pathway_aucell,pathway_aucell_enrichment,
                      geneset_aucell_tmp,pathway_aucell_tmp,pathway_aucell_enrichment_tmp,
                      pathway_enrichment,pathway_enrichment_plot,)
from ._via import pyVIA,scRNA_hematopoiesis
from ._simba import pySIMBA
from ._tosica import pyTOSICA
from ._atac import atac_concat_get_index,atac_concat_inner,atac_concat_outer
from ._batch import batch_correction
from ._cellfategenie import Fate,gene_trends,mellon_density
from ._ltnn import scLTNN,plot_origin_tesmination,find_related_gene
from ._traj import TrajInfer,fle
from ._diffusionmap import diffmap
from ._cefcon import pyCEFCON,convert_human_to_mouse_network,load_human_prior_interaction_network,mouse_hsc_nestorowa16
from ._aucell import aucell
from ._metacell import MetaCell,plot_metacells,get_obs_value
from ._mdic3 import pyMDIC3
from ._cnmf import *
from ._gptcelltype import gptcelltype,gpt4celltype,get_cluster_celltype
from ._cytotrace2 import cytotrace2
from ._gptcelltype_local import gptcelltype_local
from ._sccaf import SCCAF_assessment,plot_roc,SCCAF_optimize_all,color_long
from ._multimap import TFIDF_LSI,Wrapper,Integration,Batch
from ._scdiffusion import scDiffusion
from ._cellvote import get_cluster_celltype,CellVote
from ._deg_ct import DCT,DEG
from ._lazy_function import lazy
from ._lazy_report import generate_scRNA_report
from ._diffusionmap import diffmap
from ._cellmatch import CellOntologyMapper,download_cl
from ._scenic import SCENIC,build_correlation_network_umap_layout,add_tf_regulation,plot_grn


__all__ = [
    # Core analysis functions
    'cosg',
    'lazy',
    'aucell',
    
    # Cell type annotation
    'pySCSA',
    'MetaTiME', 
    'scanpy_lazy',
    'scanpy_cellanno_from_dict',
    'get_celltype_marker',
    'gptcelltype',
    'gpt4celltype',
    'get_cluster_celltype',
    'gptcelltype_local',
    'CellVote',
    'CellOntologyMapper',
    'download_cl',
    
    # Multi-omics integration
    'pyMOFAART',
    'pyMOFA',
    'GLUE_pair',
    'factor_exact',
    'factor_correlation',
    'get_weights',
    'glue_pair',
    'get_r2_from_hdf5',
    'pySIMBA',
    'pyTOSICA',
    'TFIDF_LSI',
    'Wrapper',
    'Integration',
    'Batch',
    
    # Trajectory and pseudotime analysis
    'pyVIA',
    'scRNA_hematopoiesis',
    'TrajInfer',
    'fle',
    'diffmap',
    'scLTNN',
    'plot_origin_tesmination',
    'find_related_gene',
    'cytotrace2',
    
    # Cell fate and development
    'Fate',
    'gene_trends',
    'mellon_density',
    
    # Network and communication analysis
    'scnocd',
    'pyCEFCON',
    'convert_human_to_mouse_network',
    'load_human_prior_interaction_network',
    'mouse_hsc_nestorowa16',
    'cpdb_network_cal',
    'cpdb_plot_network',
    'cpdb_plot_interaction',
    'cpdb_interaction_filtered',
    'cpdb_submeans_exacted',
    'cpdb_exact_target',
    'cpdb_exact_source',
    'cellphonedb_v5',
    'run_cellphonedb_v5',
    # Pathway and functional analysis
    'geneset_aucell',
    'pathway_aucell',
    'pathway_aucell_enrichment',
    'geneset_aucell_tmp',
    'pathway_aucell_tmp',
    'pathway_aucell_enrichment_tmp',
    'pathway_enrichment',
    'pathway_enrichment_plot',
    
    # Drug response analysis
    'autoResolution',
    'writeGEP',
    'Drug_Response',
    'scDiffusion',
    
    # ATAC-seq analysis
    'atac_concat_get_index',
    'atac_concat_inner',
    'atac_concat_outer',
    
    # Batch correction and preprocessing  
    'batch_correction',
    
    # Quality control and assessment
    'SCCAF_assessment',
    'plot_roc',
    'SCCAF_optimize_all',
    'color_long',
    
    # Metacells and aggregation
    'MetaCell',
    'plot_metacells',
    'get_obs_value',
    
    # Differential expression
    'DCT',
    'DEG',
    'pyMDIC3',
    
    # Additional analysis tools  
    'cnmf',
    'generate_scRNA_report',
    'SCENIC',
    'build_correlation_network_umap_layout',
    'add_tf_regulation',
    'plot_grn',
]




from typing import Sequence, Mapping, Literal, Union, Optional, Tuple, Dict
import numpy as np
import pandas as pd
import marsilea as ma
import marsilea.plotter as mp
from matplotlib.colors import Normalize, Colormap
from matplotlib.axes import Axes as _AxesSubplot
from anndata import AnnData
import functools
import operator
from typing import Any
from ._palette import palette_28, palette_56

_VarNames = Union[str, Sequence[str]]

def dotplot(
    adata: AnnData,
    var_names: Union[_VarNames, Mapping[str, _VarNames]],
    groupby: Union[str, Sequence[str]],
    *,
    use_raw: Optional[bool] = None,
    log: bool = False,
    num_categories: int = 7,
    categories_order: Optional[Sequence[str]] = None,
    expression_cutoff: float = 0.0,
    mean_only_expressed: bool = False,
    standard_scale: Optional[Literal['var', 'group']] = None,
    title: Optional[str] = None,
    colorbar_title: Optional[str] = 'Mean expression\nin group',
    size_title: Optional[str] = 'Fraction of cells\nin group (%)',
    figsize: Optional[Tuple[float, float]] = None,
    dendrogram: Union[bool, str] = False,
    gene_symbols: Optional[str] = None,
    var_group_positions: Optional[Sequence[Tuple[int, int]]] = None,
    var_group_labels: Optional[Sequence[str]] = None,
    var_group_rotation: Optional[float] = None,
    layer: Optional[str] = None,
    swap_axes: Optional[bool] = False,
    dot_color_df: Optional[pd.DataFrame] = None,
    show: Optional[bool] = None,
    save: Optional[Union[str, bool]] = None,
    ax: Optional[_AxesSubplot] = None,
    return_fig: Optional[bool] = False,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    vcenter: Optional[float] = None,
    norm: Optional[Normalize] = None,
    cmap: Union[Colormap, str, None] = 'Reds',
    dot_max: Optional[float] = None,
    dot_min: Optional[float] = None,
    smallest_dot: float = 0.0,
    fontsize: int = 12,
    preserve_dict_order: bool = False,
    **kwds,
) -> Optional[Union[Dict, 'DotPlot']]:
    r"""
    Make a dot plot of the expression values of `var_names`.
    
    For each var_name and each `groupby` category a dot is plotted.
    Each dot represents two values: mean expression within each category
    (visualized by color) and fraction of cells expressing the `var_name` in the
    category (visualized by the size of the dot).
    
    Arguments:
        adata: AnnData
            Annotated data matrix.
        var_names: str or list of str or dict
            Variables to plot.
        groupby: str or list of str
            The key of the observation grouping to consider.
        use_raw: bool, optional (default=None)
            Use `raw` attribute of `adata` if present.
        log: bool, optional (default=False)
            Whether to log-transform the data.
        num_categories: int, optional (default=7)
            Number of categories to show.
        categories_order: list of str, optional (default=None)
            Order of categories to display.
        expression_cutoff: float, optional (default=0.0)
            Expression cutoff for calculating fraction of expressing cells.
        mean_only_expressed: bool, optional (default=False)
            Whether to calculate mean only for expressing cells.
        standard_scale: {'var', 'group'} or None, optional (default=None)
            Whether to standardize data.
        title: str, optional (default=None)
            Title for the plot.
        colorbar_title: str, optional (default='Mean expression\nin group')
            Title for the color bar.
        size_title: str, optional (default='Fraction of cells\nin group (%)')
            Title for the size legend.
        figsize: tuple, optional (default=None)
            Figure size (width, height) in inches. If provided, the plot dimensions will be scaled accordingly.
        dendrogram: bool or str, optional (default=False)
            Whether to add dendrogram to the plot.
        gene_symbols: str, optional (default=None)
            Key for gene symbols in `adata.var`.
        var_group_positions: list of tuples, optional (default=None)
            Positions for variable groups.
        var_group_labels: list of str, optional (default=None)
            Labels for variable groups.
        var_group_rotation: float, optional (default=None)
            Rotation angle for variable group labels.
        layer: str, optional (default=None)
            Layer to use for expression data.
        swap_axes: bool, optional (default=False)
            Whether to swap x and y axes.
        dot_color_df: pandas.DataFrame, optional (default=None)
            DataFrame for dot colors.
        show: bool, optional (default=None)
            Whether to show the plot.
        save: str or bool, optional (default=None)
            Whether to save the plot.
        ax: matplotlib.axes.Axes, optional (default=None)
            Axes object to plot on.
        return_fig: bool, optional (default=False)
            Whether to return the figure object.
        vmin: float, optional (default=None)
            Minimum value for color scaling.
        vmax: float, optional (default=None)
            Maximum value for color scaling.
        vcenter: float, optional (default=None)
            Center value for diverging colormap.
        norm: matplotlib.colors.Normalize, optional (default=None)
            Normalization object for colors.
        cmap: str or matplotlib.colors.Colormap, optional (default='Reds')
            Colormap for the plot.
        dot_max: float, optional (default=None)
            Maximum dot size.
        dot_min: float, optional (default=None)
            Minimum dot size.
        smallest_dot: float, optional (default=0.0)
            Size of the smallest dot.
        fontsize: int, optional (default=12)
            Font size for labels and legends. Titles will be one point larger.
        preserve_dict_order: bool, optional (default=False)
            When var_names is a dictionary, whether to preserve the original dictionary order.
            If True, genes will be ordered according to the dictionary's insertion order.
            If False (default), genes will be ordered according to cell type categories.
    
    Returns:
        If `return_fig` is True, returns the figure object.
        If `show` is False, returns axes dictionary.
    """
    # Convert var_names to list if string
    original_var_names_dict = None
    if isinstance(var_names, str):
        var_names = [var_names]
    elif isinstance(var_names, Mapping):
        # Save original dictionary reference for color bar ordering
        if preserve_dict_order:
            original_var_names_dict = var_names
        
        # Get gene groups
        gene_groups = []
        var_names_list = []
        
        if preserve_dict_order:
            # Preserve the original dictionary order
            for group, genes in var_names.items():
                if isinstance(genes, str):
                    genes = [genes]
                var_names_list.extend(genes)
                gene_groups.extend([group] * len(genes))
        else:
            # Get cell type order (original behavior)
            if categories_order is not None:
                group_order = categories_order
            elif pd.api.types.is_categorical_dtype(adata.obs[groupby]):
                group_order = list(adata.obs[groupby].cat.categories)
            else:
                group_order = list(adata.obs[groupby].unique())
            
            # Order gene groups according to cell types
            for group in group_order:
                if group in var_names:
                    genes = var_names[group]
                    if isinstance(genes, str):
                        genes = [genes]
                    var_names_list.extend(genes)
                    gene_groups.extend([group] * len(genes))
            
            # Add any remaining groups that weren't in the cell types
            for group, genes in var_names.items():
                if group not in group_order:
                    if isinstance(genes, str):
                        genes = [genes]
                    var_names_list.extend(genes)
                    gene_groups.extend([group] * len(genes))
        
        var_names = var_names_list
    
    # Get expression matrix
    if use_raw and adata.raw is not None:
        matrix = adata.raw.X
        var_names_idx = [adata.raw.var_names.get_loc(name) for name in var_names]
    else:
        matrix = adata.X if layer is None else adata.layers[layer]
        var_names_idx = [adata.var_names.get_loc(name) for name in var_names]
    
    # Determine category order
    if categories_order is not None:
        cats = categories_order
    else:
        # Use the categorical order from adata if available
        if pd.api.types.is_categorical_dtype(adata.obs[groupby]):
            cats = adata.obs[groupby].cat.categories
        else:
            # If not categorical, get unique values
            cats = adata.obs[groupby].unique()
    
    # Get aggregated data with specified order
    agg = adata.obs[groupby].value_counts().reindex(cats)
    cell_counts = agg.to_numpy()
    
    # Get colors for cell types if available
    cell_colors = None
    color_dict = None
    try:
        color_key = f"{groupby}_colors"
        if color_key in adata.uns:
            colors = adata.uns[color_key]
            # Create color dictionary mapping cell types to colors
            if pd.api.types.is_categorical_dtype(adata.obs[groupby]):
                # Use categorical order for colors
                color_dict = dict(zip(adata.obs[groupby].cat.categories, colors))
            else:
                # Use unique order for colors
                unique_cats = adata.obs[groupby].unique()
                color_dict = dict(zip(unique_cats, colors[:len(unique_cats)]))
            
            # Get colors for the actual categories in the plot
            cell_colors = [color_dict.get(cat, '#CCCCCC') for cat in agg.index]
    except (KeyError, IndexError):
        cell_colors = None
        color_dict = None
    
    # Calculate mean expression and fraction of expressing cells
    means = np.zeros((len(agg), len(var_names)))
    fractions = np.zeros_like(means)
    
    for i, group in enumerate(agg.index):
        mask = adata.obs[groupby] == group
        group_matrix = matrix[mask][:, var_names_idx]
        
        # Calculate mean expression
        if mean_only_expressed:
            expressed = group_matrix > expression_cutoff
            means[i] = np.array([
                group_matrix[:, j][expressed[:, j]].mean() if expressed[:, j].any() else 0
                for j in range(group_matrix.shape[1])
            ])
        else:
            means[i] = np.mean(group_matrix, axis=0)
        
        # Calculate fraction of expressing cells
        fractions[i] = np.mean(group_matrix > expression_cutoff, axis=0)
    
    # Scale if requested
    if standard_scale == 'group':
        means = (means - means.min(axis=1, keepdims=True)) / (means.max(axis=1, keepdims=True) - means.min(axis=1, keepdims=True))
    elif standard_scale == 'var':
        means = (means - means.min(axis=0)) / (means.max(axis=0) - means.min(axis=0))
    
    # Handle dot size limits
    if dot_max is not None:
        fractions = np.minimum(fractions, dot_max)
    if dot_min is not None:
        fractions = np.maximum(fractions, dot_min)
    
    # Scale dot sizes to account for smallest_dot
    if smallest_dot > 0:
        fractions = smallest_dot + (1 - smallest_dot) * fractions
    
    # Create the plot
    h, w = means.shape
    
    # Calculate dimensions based on figsize if provided
    if figsize is not None:
        # Use figsize to determine height and width
        # Adjust for the number of rows and columns to maintain aspect ratio
        base_height = figsize[1] * 0.7  # Use 70% of figsize height for main plot
        base_width = figsize[0] * 0.7   # Use 70% of figsize width for main plot
        
        # Scale based on data dimensions
        height = base_height * (h / max(h, w))
        width = base_width * (w / max(h, w))
    else:
        # Default behavior
        height = h / 3
        width = w / 3
    
    
    # Create SizedHeatmap
    m = ma.SizedHeatmap(
        size=fractions,
        color=means,
        cluster_data=fractions if dendrogram else None,
        height=height,
        width=width,
        edgecolor="lightgray",
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        #norm=norm,
        size_legend_kws=dict(
            colors="#c2c2c2",
            title=size_title,
            labels=[f"{int(x*100)}%" for x in [0.2, 0.4, 0.6, 0.8, 1.0]],
            show_at=[0.2, 0.4, 0.6, 0.8, 1.0],
            fontsize=fontsize,
            ncol=3,
            title_fontproperties={"size": fontsize + 1, "weight": 100}
        ),
        color_legend_kws=dict(
            title=colorbar_title,
            fontsize=fontsize,
            orientation="horizontal",
            title_fontproperties={"size": fontsize + 1, "weight": 100}
        ),
    )
    
    # Add labels
    m.add_top(mp.Labels(var_names, fontsize=fontsize), pad=0.1)
    
    # Group genes if var_names was a dictionary
    if 'gene_groups' in locals():
        # Get colors for gene groups
        try:
            # Use the same color_dict that was created for cell types
            if color_dict is not None:
                # Get unique groups and check which ones are not in color_dict
                unique_groups = list(dict.fromkeys(gene_groups))
                missing_groups = [g for g in unique_groups if g not in color_dict]
                
                # If there are missing groups, add colors from palette
                if missing_groups:
                    if len(palette_28) >= len(color_dict) + len(missing_groups):
                        extra_colors = palette_28[len(color_dict):len(color_dict) + len(missing_groups)]
                    else:
                        extra_colors = palette_56[len(color_dict):len(color_dict) + len(missing_groups)]
                    color_dict.update(dict(zip(missing_groups, extra_colors)))
            else:
                # If no colors found in uns, use default palette
                unique_groups = list(dict.fromkeys(gene_groups))
                if len(unique_groups) <= 28:
                    palette = palette_28
                else:
                    palette = palette_56
                color_dict = dict(zip(unique_groups, palette[:len(unique_groups)]))
        except (KeyError, AttributeError):
            # If colors not found in uns, use default palette
            unique_groups = list(dict.fromkeys(gene_groups))
            if len(unique_groups) <= 28:
                palette = palette_28
            else:
                palette = palette_56
            color_dict = dict(zip(unique_groups, palette[:len(unique_groups)]))
        
        # Add color bars with matching order
        # Add group labels
        # Only show used colors in legend and increase group spacing
        if preserve_dict_order and original_var_names_dict is not None:
            # When preserving dict order, use the original dictionary key order
            used_groups = list(original_var_names_dict.keys())
        else:
            # Use the order as they appear in gene_groups
            used_groups = list(dict.fromkeys(gene_groups))
        
        used_color_dict = {k: color_dict[k] for k in used_groups}
        m.add_top(
            mp.Colors(gene_groups, palette=used_color_dict),
            pad=0.1,
            size=0.15,
        )
        # Add group labels with increased spacing
        m.group_cols(gene_groups,order=used_groups)
    
    # Add cell type colors if available
    if color_dict is not None:
        # Add color bar using the properly created color_dict
        m.add_left(
            mp.Colors(agg.index, palette=color_dict),
            size=0.15,
            pad=0.1,
            legend=False,
        )
    
    # Add cell type labels
    m.add_left(mp.Labels(agg.index, align="right", fontsize=fontsize), pad=0.1)
    
    # Add cell counts
    m.add_right(
        mp.Numbers(
            cell_counts,
            color="#EEB76B",
            label="Count",
            label_props={'size': fontsize},
            props={'size': fontsize},
            show_value=False
        ),
        size=0.5,
        pad=0.1,
    )
    
    # Add dendrogram if requested
    if dendrogram:
        m.add_dendrogram("right", pad=0.1)
    
    # Add legends
    m.add_legends(box_padding=2)
    
    # Render the plot
    fig = m.render()
    
    if return_fig:
        return fig
    elif not show:
        return m
    return None

def rank_genes_groups_df(
    adata: AnnData,
    group: str,
    key: str = "rank_genes_groups",
    gene_symbols: str | None = None,
    log2fc_min: float | None = None,
) -> pd.DataFrame:
    """Return a DataFrame with the results of rank_genes_groups."""
    d = pd.DataFrame()
    for k in ['names', 'scores', 'logfoldchanges', 'pvals', 'pvals_adj']:
        if k in adata.uns[key]:
            d[k] = pd.DataFrame(adata.uns[key][k])[group]
    
    if log2fc_min is not None:
        d = d[d['logfoldchanges'].abs() > log2fc_min]
    
    return d

def _get_values_to_plot(
    adata: AnnData,
    values_to_plot: str,
    var_names: Sequence[str],
    key: str = 'rank_genes_groups',
    gene_symbols: str | None = None,
) -> pd.DataFrame:
    """Get values to plot from rank_genes_groups results."""
    if values_to_plot not in adata.uns[key]:
        raise ValueError(
            f'The key {values_to_plot} is not available in adata.uns["{key}"]'
        )
    
    # Get the values for each group
    values = pd.DataFrame(adata.uns[key][values_to_plot])
    values.index = pd.DataFrame(adata.uns[key]['names']).iloc[:, 0]
    values = values.loc[var_names]
    
    return values

def rank_genes_groups_dotplot(
    adata: AnnData,
    plot_type: str = "dotplot",
    *,
    groups: Optional[Union[str, Sequence[str]]] = None,
    n_genes: Optional[int] = None,
    groupby: Optional[str] = None,
    values_to_plot: Optional[str] = None,
    var_names: Optional[Union[Sequence[str], Mapping[str, Sequence[str]]]] = None,
    min_logfoldchange: Optional[float] = None,
    key: Optional[str] = None,
    show: Optional[bool] = None,
    save: Optional[bool] = None,
    return_fig: bool = False,
    gene_symbols: Optional[str] = None,
    **kwds: Any,
) -> Optional[Union[Dict, Any]]:
    """
    Create a dot plot from rank_genes_groups results.
    
    Parameters
    ----------
    adata : AnnData
        Annotated data matrix.
    plot_type : str
        Currently only 'dotplot' is supported.
    groups : str or list of str, optional
        Groups to include in the plot.
    n_genes : int, optional
        Number of genes to include in the plot.
    groupby : str, optional
        Key in `adata.obs` to group by.
    values_to_plot : str, optional
        Key in rank_genes_groups results to plot (e.g. 'logfoldchanges', 'scores').
    var_names : str or list of str or dict, optional
        Variables to include in the plot. Can be:
        - A list of gene names: ['gene1', 'gene2', ...]
        - A dictionary mapping group names to gene lists: {'group1': ['gene1', 'gene2'], 'group2': ['gene3', 'gene4']}
        When a dictionary is provided, genes will be grouped and labeled accordingly in the plot.
    min_logfoldchange : float, optional
        Minimum log fold change to include in the plot.
    key : str, optional
        Key in `adata.uns` to use for rank_genes_groups results.
    show : bool, optional
        Whether to show the plot.
    save : bool, optional
        Whether to save the plot.
    return_fig : bool
        Whether to return the figure object.
    gene_symbols : str, optional
        Key for gene symbols in `adata.var`.
    **kwds : dict
        Additional keyword arguments to pass to dotplot.
    
    Returns
    -------
    If `return_fig` is True, returns the figure object.
    If `show` is False, returns axes dictionary.
    
    Examples
    --------
    >>> # Basic usage with top genes
    >>> sc.pl.rank_genes_groups_dotplot(adata, n_genes=5)
    
    >>> # Using logfoldchanges for coloring
    >>> sc.pl.rank_genes_groups_dotplot(adata, n_genes=5, values_to_plot='logfoldchanges')
    
    >>> # Grouping genes manually
    >>> gene_groups = {
    ...     'Group1': ['gene1', 'gene2'],
    ...     'Group2': ['gene3', 'gene4']
    ... }
    >>> sc.pl.rank_genes_groups_dotplot(adata, var_names=gene_groups)
    """
    if plot_type != "dotplot":
        raise ValueError("Only 'dotplot' is currently supported")
        
    if var_names is not None and n_genes is not None:
        msg = (
            "The arguments n_genes and var_names are mutually exclusive. Please "
            "select only one."
        )
        raise ValueError(msg)

    if key is None:
        key = "rank_genes_groups"

    if groupby is None:
        groupby = str(adata.uns[key]["params"]["groupby"])
    group_names = adata.uns[key]["names"].dtype.names if groups is None else groups

    if var_names is not None:
        if isinstance(var_names, Mapping):
            # get a single list of all gene names in the dictionary
            var_names_list = functools.reduce(
                operator.iadd, [list(x) for x in var_names.values()], []
            )
        elif isinstance(var_names, str):
            var_names_list = [var_names]
        else:
            var_names_list = var_names
    else:
        # set n_genes = 10 as default when none of the options is given
        if n_genes is None:
            n_genes = 10

        # dict in which each group is the key and the n_genes are the values
        var_names = {}
        var_names_list = []
        for group in group_names:
            df = rank_genes_groups_df(
                adata,
                group,
                key=key,
                gene_symbols=gene_symbols,
                log2fc_min=min_logfoldchange,
            )

            if gene_symbols is not None:
                df["names"] = df[gene_symbols]

            genes_list = df.names[df.names.notnull()].tolist()

            if len(genes_list) == 0:
                print(f"Warning: No genes found for group {group}")
                continue
            genes_list = genes_list[n_genes:] if n_genes < 0 else genes_list[:n_genes]
            var_names[group] = genes_list
            var_names_list.extend(genes_list)

    # by default add dendrogram to plots
    kwds.setdefault("dendrogram", True)

    # Get values to plot if specified
    title = None
    values_df = None
    if values_to_plot is not None:
        values_df = _get_values_to_plot(
            adata,
            values_to_plot,
            var_names_list,
            key=key,
            gene_symbols=gene_symbols,
        )
        title = values_to_plot
        if values_to_plot == "logfoldchanges":
            title = "log fold change"
        else:
            title = values_to_plot.replace("_", " ").replace("pvals", "p-value")

    # Create the plot
    _pl = dotplot(
        adata,
        var_names,
        groupby,
        dot_color_df=values_df,
        return_fig=True,
        gene_symbols=gene_symbols,
        preserve_dict_order=True,
        **kwds,
    )
    
    if title is not None and "colorbar_title" not in kwds:
        _pl.legend(colorbar_title=title)
    
    if return_fig:
        return _pl
    elif not show:
        return _pl
    return None


import pandas as pd
import numpy as np

import scipy.sparse as sp
from scipy.ndimage import gaussian_filter
import scanpy as sc
from tqdm import trange, tqdm

import multiprocessing as mp
from .utils import pre_process, make_image, gene_img_flatten, minmax_normalize, gau_filter_for_single_gene



def prepare_for_PI(adata, grid_size=20, percentage=0.1, platform="visium"):    
    selected_gene_idxs, postcount = pre_process(adata, percentage, var_stabilization = False)

    if platform=="visium" or platform=="ST":
        try:
            locates = adata.obs[["array_row","array_col"]]
            locates = locates.values if isinstance(locates, pd.DataFrame) else locates
        except:
            locates = adata.obsm["spatial"]
            locates = locates.values if isinstance(locates, pd.DataFrame) else locates
        if np.min(locates) == 0:
            locates += 1
        _, image_idx = make_image(postcount[0], locates, platform, get_image_idx = True, grid_size=grid_size)
        adata = adata[:, selected_gene_idxs]
        sc.pp.filter_genes(adata, min_cells=3)
        adata.obs['image_idx_1d'] = image_idx

    else:
        locates = adata.obsm["spatial"].astype(float)
        locates = locates.values if isinstance(locates, pd.DataFrame) else locates
        _, shape = make_image(postcount[0], locates, platform, grid_size=grid_size)
        assert shape[0]>1 and shape[1]>1, f"Gene image size is {shape[0]} * {shape[1]} after interpolation. Please set a smaller grid size!!"
        print (f"Spatial gene expression is interpolated into images of size [{shape[0]} * {shape[1]}]")
        adata = adata[:, selected_gene_idxs]
        sc.pp.filter_genes(adata, min_cells=50)
        adata.uns['shape'] = shape
    adata.uns['grid_size'] = grid_size
    adata.uns['locates'] = locates
    return adata


def minmax_scaler(adata,layer='counts'):
    if layer=='X' or layer=='raw':
        if sp.issparse(adata.X):
            data = adata.X.A.T
        else:
            data = adata.X.T      
    else:
        if sp.issparse(adata.layers[layer]):
            data = adata.layers[layer].A.T
        else:
            data = adata.layers[layer].T   

    print('\nNormalize each geneing...')
    nor_counts = data.copy().T
    _nor_maxdata = np.max(nor_counts, 0)
    _nor_mindata = np.min(nor_counts, 0)
    nor_counts = (nor_counts - _nor_mindata) / (_nor_maxdata - _nor_mindata)
    adata.uns['nor_counts'] = nor_counts.T
    return adata


def gau_filter_for_single_gene(arglist):
    gene_data, locates, platform, image_idx_1d = arglist

    I,_ = make_image(gene_data, locates, platform) 
    I = gaussian_filter(I, sigma = 1, truncate = 2)
    if platform=="visium":
        I_1d = I.T.flatten()
        output = I_1d[image_idx_1d-1]
    else:
        output = I.flatten()
    return output


def gau_filter(adata, platform="visium", multiprocess=False):
    gene_data = adata.uns['nor_counts']
    locates = adata.uns['locates']
    N_gene = len(gene_data)

    print('\nGaussian filtering...')
    if platform=="visium":
        image_idx_1dd = adata.obs['image_idx_1d'].astype(int).values

    def sel_data():  # data generater
        for gene_i in range(N_gene):
            if platform=="visium":
                yield [gene_data[gene_i], locates, platform, image_idx_1dd]
            else:
                yield [gene_data[gene_i], locates, platform, '']
                
    if multiprocess:
        num_cores = int(mp.cpu_count() / 2)         # default core is half of total
        with mp.Pool(processes=num_cores) as pool:
            gau_fea = list(tqdm(pool.imap(gau_filter_for_single_gene, sel_data()), total=N_gene))
    else:
        gau_fea = list(tqdm(map(gau_filter_for_single_gene, sel_data()), total=N_gene))
        
    adata.uns['gau_fea'] = np.array(gau_fea, dtype=np.float64) 
    return adata


def _iget_binary(arglists):
    import cv2
    fig1, locates, platform, method, r1 = arglists

    if platform=="visium":
        Im, _ = make_image(fig1, locates, platform)
        if method == "iterative":    
            m, n = Im.shape       
            zd = float(np.max(Im))
            zx = float(np.min(Im))
            Th = float((zd+zx))/2
            while True:
                S0 = 0.0; n0 = 0.0; S1 = 0.0; n1 = 0.0
                flag = Im >= Th
                S1 = Im[flag].sum()
                n1 = flag.sum()
                S0 = Im[~flag].sum()
                n0 = (~flag).sum()
                T0 = S0/n0; T1 = S1/n1
                if abs(Th - ((T0 + T1)/2)) < 0.0001:
                    break
                else:
                    Th = (T0 + T1)/2 
                        
        elif method == "otsu":
            thres_list = np.arange(0.01,0.995,0.025)
            temp_std = np.zeros(thres_list.shape)
            for iii in range(len(thres_list)):
                temp_thres = thres_list[iii]
                q1 = fig1 > temp_thres
                b1 = fig1 <= temp_thres
                qv = r1[q1]
                bv = r1[b1]
                if len(qv) >= len(r1) * 0.15:
                    temp_std[iii] = (len(qv) * np.std(qv) + len(bv) * np.std(bv)) / len(fig1)
                else:
                    temp_std[iii] = 1e4
            Th = thres_list[temp_std == np.min(temp_std)]
    #--------------------------------------------------------------------------        
    else:               
        if method == "iterative":        
            zd = float(np.nanmax(fig1))
            zx = float(np.nanmin(fig1))
            Th = float((zd+zx))/2
            while True:                       
                S1 = np.sum(fig1[fig1>=Th])
                n1 = len(fig1[fig1>=Th])
                S0 = np.sum(fig1[fig1<Th])
                n0 = len(fig1[fig1<Th])
                T0 = S0/n0; T1 = S1/n1
                if abs(Th - ((T0 + T1)/2)) < 0.0001:
                    break
                else:
                    Th = (T0 + T1)/2 

        elif method == "otsu":
            # for ii in trange(len(gene_data)):
            img = fig1.reshape(locates)
            Th2, a_img = cv2.threshold(img.astype(np.uint8), 0, 255, cv2.THRESH_OTSU) 

    return fig1 >= Th


def get_binary(adata, platform="visium", method = "iterative", multiprocess=False):
    gene_data = adata.uns['gau_fea']
    if sp.issparse(adata.X):
        raw_gene_data = adata.X.A.T
    else:
        raw_gene_data = adata.X.T

    if platform=="visium":
        locates = adata.uns['locates']
    else:
        locates = adata.uns['shape']
    
    print('\nBinary segmentation for each gene:')
    
    N_gene = len(gene_data)
    def sel_data():  # data generater
        for gene_i in range(N_gene):
            yield [gene_data[gene_i, :], locates, platform, method, raw_gene_data[gene_i,:]]

    if multiprocess:
        num_cores = int(mp.cpu_count() / 2)         # default core is half of total
        with mp.Pool(processes=num_cores) as pool:
            output = list(tqdm(pool.imap(_iget_binary, sel_data()), total=N_gene))
    else:
        output = list(tqdm(map(_iget_binary, sel_data()), total=N_gene))
    output = np.array(output, dtype=np.float64) + 0.0
    adata.uns['binary_image'] = output
    return adata


def get_sub(adata, kernel_size = 5, platform="visium",del_rate = 0.01): 
    import cv2
    from skimage.measure import label
    gene_data = adata.uns['binary_image']
    locates = adata.uns['locates']
        
    print('\nSpliting subregions for each gene:')
    #--------------------------------------------------------------------------
    if platform=="visium":
        image_idx_1d = adata.obs['image_idx_1d']
        output = np.zeros(gene_data.shape)
        del_index = np.ones(gene_data.shape[0])
        for i in tqdm(range(len(gene_data))):
            temp_data = gene_data[i, :]
            temp_i, _ = make_image(temp_data, locates)      
            kernel = np.ones((kernel_size,kernel_size), np.uint8)
            temp_i = cv2.morphologyEx(temp_i, cv2.MORPH_CLOSE, kernel) # close
            region_label = label(temp_i)
            T = np.zeros(region_label.shape)
            classes = np.max(np.unique(region_label)) + 1      
            len_list = np.zeros(classes)     
            for j in range(classes):
                len_list[j] = len(region_label[region_label == j])
            cond = len_list >= gene_data.shape[1] * 0.01        
            if len(np.where(cond[1:] == True)[0]) == 0:
                del_index[i] = 0
            indexes = np.where(cond == True)[0]       
            for j in range(len(indexes)):
                tar_num = indexes[j]
                tar_locs = region_label == tar_num
                T[tar_locs] = j
            targe_image = T * (temp_i > 0)
            classes_n = np.max(np.unique(targe_image)).astype(int) + 1       
            len_list_n = np.zeros(classes_n)        
            for j in range(classes_n):
                len_list_n[j] = len(targe_image[targe_image == j])            
            if len(len_list_n) > 1:                        
                if np.max(len_list_n[1:]) < gene_data.shape[1] * del_rate:
                    del_index[i] = 0
            else:
                del_index[i] = 0
            output[i, :] = gene_img_flatten(targe_image, image_idx_1d.values)      
    #--------------------------------------------------------------------------
    else:
        output = np.zeros((gene_data.shape[0], adata.uns['shape'][0]*adata.uns['shape'][1]))
        del_index = np.ones(gene_data.shape[0])
        for i in tqdm(range(len(gene_data))):
            temp_data = gene_data[i, :]
            temp_i = temp_data.reshape(adata.uns['shape'])     
            kernel = np.ones((kernel_size,kernel_size), np.uint8)
            temp_i = cv2.morphologyEx(temp_i, cv2.MORPH_CLOSE, kernel)
            region_label = label(temp_i)
            T = np.zeros(region_label.shape)
            classes = np.max(region_label) + 1      
            len_list = np.zeros(classes)
            for j in range(classes):
                len_list[j] = len(region_label[region_label == j])
            cond = len_list >= gene_data.shape[1] * 0.002        
            if len(np.where(cond[1:] == True)[0]) == 0:
                del_index[i] = 0
            indexes = np.where(cond == True)[0]       
            for j in range(len(indexes)):
                tar_num = indexes[j]
                tar_locs = region_label == tar_num
                T[tar_locs] = j
            targe_image = T * (temp_i > 0)
            classes_n = np.max(np.unique(targe_image)).astype(int) + 1       
            len_list_n = np.zeros(classes_n)        
            for j in range(classes_n):
                len_list_n[j] = len(targe_image[targe_image == j])            
            if len(len_list_n) > 1:                        
                if np.max(len_list_n[1:]) < gene_data.shape[1] * del_rate:
                    del_index[i] = 0
            else:
                del_index[i] = 0
            output[i, :] = targe_image.flatten()
    #--------------------------------------------------------------------------
    adata.uns['subregions'] = output
    adata.uns['del_index'] = del_index.astype(int)  
    return adata


def cal_prost_index(adata, platform="visium"):
    data = adata.uns['nor_counts']
    subregions = adata.uns['subregions']
    del_idx = adata.uns['del_index']
    
    print('\nComputing PROST Index for each gene:')
    #--------------------------------------------------------------------------
    if platform=="visium":
        SEP = np.zeros(len(data))
        SIG = np.zeros(len(data))
        region_number = np.zeros(len(data))
        
        for i in tqdm(range(len(data))): 
            temp_raw = data[i, :]
            temp_label = subregions[i, :]
            back_value = temp_raw[temp_label == 0]
            back_value = back_value[back_value > 0]
            if back_value.size == 0:
                back_value = 0  
            class_mean = np.zeros(max(np.unique(temp_label)).astype(int) + 1)
            class_var = np.zeros(max(np.unique(temp_label)).astype(int) + 1)
            class_std = np.zeros(max(np.unique(temp_label)).astype(int) + 1)
            class_len = np.zeros(max(np.unique(temp_label)).astype(int) + 1)
         
            for ii in range(max(np.unique(temp_label)).astype(int) + 1):
                Temp = temp_raw[temp_label == ii]
                if Temp.size == 0:
                    class_value = 0
                else:
                    class_value = Temp
                class_mean[ii] = np.mean(class_value)
                class_var[ii] = np.var(class_value)
                class_std[ii] = np.std(class_value)
                if isinstance(class_value, int):
                    if class_value == 0:
                        class_len[ii] = 0
                    else:              
                        class_len[ii] = len(class_value) - 1               
                else:
                    if class_value.size == 0:
                        class_len[ii] = 0
                    else:              
                        class_len[ii] = len(class_value) - 1
                        
            target_class = np.where(class_mean > 0)[0]
            class_mean = class_mean[target_class]
            class_std = class_std[target_class]
            class_var = class_var[target_class]
            class_len = class_len[target_class]
            
            # Calculate Separability and Significance
            SEP[i] = 1 - sum((class_len * class_var)) / ((len(temp_raw)-1) * np.var(temp_raw))
            SIG[i] = (np.mean(class_mean) - np.mean(back_value)) / sum(class_std / class_mean) 
            region_number[i] = len(class_len)
            del class_mean, class_var, class_len, class_std
                     
        # Pattern Index    
        PI = minmax_normalize(SEP) * minmax_normalize(SIG)
        PI = PI * del_idx   
        adata.var["SEP"] = SEP
        adata.var["SIG"] = SIG
        adata.var["PI"] = PI
    #--------------------------------------------------------------------------
    else:
        locates = adata.uns['locates']
        SEP = np.zeros(len(data))
        SIG = np.zeros(len(data))
        for i in tqdm(range(len(data))): 
            temp_raw = data[i, :]
            temp_img,_ = make_image(temp_raw, locates, platform)
            temp_raw = temp_img.flatten()
            temp_label = subregions[i, :]
            back_value = temp_raw[temp_label == 0]
            back_value = back_value[back_value > 0]
            
            if back_value.size == 0:
                back_value = 0  
            class_mean = np.zeros(max(np.unique(temp_label)).astype(int) + 1)
            class_var = np.zeros(max(np.unique(temp_label)).astype(int) + 1)
            class_std = np.zeros(max(np.unique(temp_label)).astype(int) + 1)
            class_len = np.zeros(max(np.unique(temp_label)).astype(int) + 1)
         
            for ii in range(max(np.unique(temp_label)).astype(int) + 1):
                Temp = temp_raw[temp_label == ii]
                if Temp.size == 0:
                    class_value = 0
                else:
                    class_value = Temp
                class_mean[ii] = np.nanmean(class_value)
                class_var[ii] = np.nanvar(class_value)
                class_std[ii] = np.nanstd(class_value)
                if isinstance(class_value, int):
                    if class_value == 0:
                        class_len[ii] = 0
                    else:              
                        class_len[ii] = len(class_value) - 1               
                else:
                    if class_value.size == 0:
                        class_len[ii] = 0
                    else:              
                        class_len[ii] = len(class_value) - 1
                        
            target_class = np.where(class_mean > 0)[0]
            class_mean = class_mean[target_class]
            class_std = class_std[target_class]
            class_var = class_var[target_class]
            class_len = class_len[target_class]
            
            # Calculate Separability and Significance
            SEP[i] = 1 - sum((class_len * class_var)) / ((len(temp_raw)-1) * np.nanvar(temp_raw))
            SIG[i] = (np.mean(class_mean) - np.mean(back_value)) / sum(class_std / class_mean)      
            del class_mean, class_var, class_len, class_std
            
        # Pattern Index    
        PI = minmax_normalize(SEP) * minmax_normalize(SIG)
        PI = PI * del_idx
        adata.var["SEP"] = SEP
        adata.var["SIG"] = SIG
        adata.var["PI"] = PI
        
        adata.uns['shape'] = []
    #--------------------------------------------------------------------------
    adata.uns['nor_counts'] = []
    adata.uns['binary_image'] = []
    adata.uns['subregions'] = []
    adata.uns['del_index'] = []
    
    return adata
